#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = ( 'JmsNode', )

from ..constants import ( JMS_VERSION_HALO_1, JMS_VERSION_HALO_2_8210, )

class JmsNode:
    __slots__ = (
        "name",
        "first_child", "sibling_index", "parent_index",
        "rot_i", "rot_j", "rot_k", "rot_w",
        "pos_x", "pos_y", "pos_z",
        )
    def __init__(self, name="", first_child=-1, sibling_index=-1,
                 rot_i=0.0, rot_j=0.0, rot_k=0.0, rot_w=1.0,
                 pos_x=0.0, pos_y=0.0, pos_z=0.0, parent_index=-1):
        self.name = name
        self.sibling_index = sibling_index
        self.first_child = first_child
        self.rot_i = rot_i
        self.rot_j = rot_j
        self.rot_k = rot_k
        self.rot_w = rot_w
        self.pos_x = pos_x
        self.pos_y = pos_y
        self.pos_z = pos_z
        self.parent_index = parent_index

    def __repr__(self):
        return """JmsNode(name=%s,
    first_child=%s, sibling_index=%s,
    i=%s, j=%s, k=%s, w=%s,
    x=%s, y=%s, z=%s
)""" % (self.name, self.first_child, self.sibling_index,
        self.rot_i, self.rot_j, self.rot_k, self.rot_w,
        self.pos_x, self.pos_y, self.pos_z)

    def __eq__(self, other):
        if not isinstance(other, JmsNode):
            return False
        elif self.name != other.name:
            return False
        elif self.first_child != other.first_child:
            return False
        elif self.sibling_index != other.sibling_index:
            return False
        elif (abs(self.rot_i - other.rot_i) > 0.00001 or
              abs(self.rot_j - other.rot_j) > 0.00001 or
              abs(self.rot_k - other.rot_k) > 0.00001 or
              abs(self.rot_w - other.rot_w) > 0.00001):
            return False
        elif (abs(self.pos_x - other.pos_x) > 0.00001 or
              abs(self.pos_y - other.pos_y) > 0.00001 or
              abs(self.pos_z - other.pos_z) > 0.00001):
            return False
        return True

    def is_node_hierarchy_equal(self, other):
        if not isinstance(other, JmsNode):
            return False
        elif self.name != other.name:
            return False
        elif self.first_child != other.first_child:
            return False
        elif self.sibling_index != other.sibling_index:
            return False
        return True

    @classmethod
    def setup_node_hierarchy(cls, nodes, jms_version=JMS_VERSION_HALO_1):
        if jms_version == JMS_VERSION_HALO_1:
            # Halo 1
            parented_nodes = set()
            # setup the parent node hierarchy
            for parent_idx in range(len(nodes)):
                node = nodes[parent_idx]
                if node.first_child > 0:
                    sib_idx = node.first_child
                    seen_nodes = set()
                    while sib_idx >= 0:
                        if (sib_idx in seen_nodes or sib_idx == parent_idx or
                            sib_idx >= len(nodes)):
                            break
                        seen_nodes.add(sib_idx)
                        parented_nodes.add(sib_idx)
                        sib_node = nodes[sib_idx]
                        sib_node.parent_index = parent_idx
                        sib_idx = sib_node.sibling_index
        elif jms_version == JMS_VERSION_HALO_2_8210:
            # Halo 2
            pass



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.vehi import *
from .obje import *

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=1)

vehi_body = Struct("tagdata",
    obje_attrs,
    unit_attrs,
    vehi_attrs,
    SIZE=1008,
    )


def get():
    return vehi_def

vehi_def = TagDef("vehi",
    blam_header('vehi'),
    vehi_body,

    ext=".vehicle", endian=">", tag_cls=ObjeTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .shdr import *
from .objs.shdr import ShdrTag
from supyr_struct.defs.tag_def import TagDef

noise_map = Struct("noise_map",
    FlUInt16("unknown0", VISIBLE=False),
    FlUInt16("unknown1", VISIBLE=False),
    float_sec("animation_period"),
    QStruct("animation_direction", INCLUDE=ijk_float),
    Float("noise_map_scale"),
    dependency("noise_map", "bitm"),
    Pad(32),
    )

spla_attrs = Struct("spla_attrs",
    Pad(4),
    #Intensity
    Struct("intensity",
        SEnum16("source", *function_outputs),
        Pad(2),
        Float("exponent"),
        ),

    #Offset
    Struct("offset",
        SEnum16("source", *function_outputs),
        Pad(2),
        float_wu("amount"),
        Float("exponent"),
        ),

    Pad(32),

    #Color
    Struct("color",
        float_zero_to_one("perpendicular_brightness"),
        QStruct("perpendicular_tint_color", INCLUDE=rgb_float),
        float_zero_to_one("parallel_brightness"),
        QStruct("parallel_tint_color", INCLUDE=rgb_float),
        SEnum16("tint_color_source", *function_names),
        ),

    Pad(58),
    #Primary Noise Map
    Struct("primary_noise_map", INCLUDE=noise_map),

    #Secondary Noise Map
    Struct("secondary_noise_map", INCLUDE=noise_map),
    SIZE=292
    )

spla_body = Struct("tagdata",
    shdr_attrs,
    spla_attrs,
    SIZE=332,
    )


def get():
    return spla_def

spla_def = TagDef("spla",
    blam_header('spla'),
    spla_body,

    ext=".shader_transparent_plasma", endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.antr import *
from supyr_struct.util import desc_variant

antr_body = desc_variant(
    antr_body,
    ("animations", reflexive(
        "animations", animation_desc, 2048, DYN_NAME_PATH=".name"))
    )

def get():
    return antr_def

antr_def = TagDef("antr",
    blam_header('antr', 4),
    antr_body,

    ext=".model_animations", endian=">", tag_cls=AntrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

'''
    Adapted from source files located here
    https://github.com/Halogen002/Flare-Qt

    My thanks go to Halogen002 for providing me with
    the information I needed to write this definition.
    I extended it to include xbox gametypes as well
'''

import binascii

from struct import unpack

from reclaimer.misc.defs.objs.xboxsave import XboxSaveTag
from reclaimer.field_types import FieldType
from supyr_struct.buffer import BytearrayBuffer, get_rawdata
from supyr_struct.tag import Tag

CE_CRC32_OFF = 0x98
PC_CRC32_OFF = 0xD8

HALO_SIGKEY = (b'\x1F\x71\xDE\x93\xD5\x2A\xAD\xB1'+
               b'\x94\x46\xD7\x49\x4F\x73\x11\x58')

class PcGametypeTag(XboxSaveTag):

    sigkey     = HALO_SIGKEY
    data_start = 0
    data_end   = 104

    is_xbox    = False
    is_powerpc = False

    def calc_crc32(self, buffer=None, offset=-1):
        '''Returns the crc32 checksum of the data in 'buffer' up
        to the 'offset' specified. If offset is not specified, the
        entire buffer is used. If buffer is not specified, the tag
        will be written and the returned buffer will be used.'''
        if buffer is None:
            buffer = self.data.serialize(buffer=BytearrayBuffer())
        return 0xFFFFFFFF - (binascii.crc32(buffer[:offset]) & 0xFFFFFFFF)

    def validate_checksum(self, buffer=None, offset=-1, end='<'):
        '''Returns True if the checksum of the gametype is valid'''
        checksum = unpack(end+'I', buffer[offset:offset+4])[0]
        return self.calc_crc32(buffer, offset) == checksum

    def parse(self, **kwargs):
        ''''''
        if kwargs.get('filepath') is None and kwargs.get('rawdata') is None:
            kwargs['filepath'] = self.filepath

        rawdata = get_rawdata(**kwargs)
        if rawdata:
            rawdata = kwargs['rawdata'] = BytearrayBuffer(rawdata)
        if 'filepath' in kwargs:
            del kwargs['filepath']

        if rawdata is not None:
            is_ce = self.validate_checksum(rawdata, CE_CRC32_OFF)
            is_pc = self.validate_checksum(rawdata, PC_CRC32_OFF)
            #if the checksum doesnt check out for either PC or CE,
            #see if the big endian version of the checksum checks out
            if not(is_ce or is_pc):
                if self.validate_checksum(rawdata, PC_CRC32_OFF, '>'):
                    #turns out the gametype is big endian, who woulda thought?
                    self.is_powerpc = is_pc = True

            self.is_xbox = not(is_ce or is_pc)

            #if the gametype isnt a valid PC gametype, make it a hybrid of both
            if is_ce and not is_pc:
                #copy the checksum to the PC Halo specific location
                rawdata[0x94:0x9C] = rawdata[0xD4:0xDC]
                #copy the gametype settings to the PC Halo specific location
                rawdata[0x7C:0x94] = rawdata[0x9C:0xB4]
        else:
            self.is_xbox = False
            self.is_powerpc = False

        #make sure to force all the fields to read normal endianness
        #after trying to read the tag, even if there is an exception
        try:
            if self.is_powerpc:
                FieldType.force_big()
            return XboxSaveTag.parse(self, **kwargs)
        finally:
            if self.is_powerpc:
                FieldType.force_normal()

    def xbox_sign(self, rawdata=None, authkey=None):
        if rawdata is None:
            rawdata = self.data.serialize(buffer=BytearrayBuffer()
                                          )[:self.data_end]

        hmac_sig = self.calc_hmac_sig(rawdata, authkey)
        self.data.gametype_footer.hmac_sig = hmac_sig

    def serialize(self, **kwargs):
        '''Writes this tag to the set path like normal, but makes
        sure to calculate and set the checksums before doing so.'''
        try:
            if self.is_powerpc:
                FieldType.force_big()
            if self.is_xbox:
                #calculate the xbox checksum
                self.xbox_sign()
            else:
                #calculate the pc/ce checksum
                footer = self.data.gametype_footer
                footer.crc_32 = self.calc_crc32(None, CE_CRC32_OFF)

                footer.hybrid_settings = BytearrayBuffer()
                self.data.gametype_settings.serialize(buffer=footer.\
                                                      hybrid_settings)

                footer.crc_32_ce = self.calc_crc32(None, PC_CRC32_OFF)
            return Tag.serialize(self, **kwargs)
        finally:
            if self.is_powerpc:
                FieldType.force_normal()



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DeadCanadian
# 	named a few things
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


jmrq_sandbox_text_value_pair_text_value_pair = Struct("text_value_pair",
    Bool8("flags",
        "default",
        "unchanged",
        "_2",
        "_3",
        ),
    SEnum8("expected_value_type", *sily_text_value_pair_expected_value_type),
    SInt16("unknown", VISIBLE=False),
    SInt32("int_value"),
    h3_string_id("ref_name"),
    h3_string_id("name"),
    h3_string_id("description"),
    ENDIAN=">", SIZE=20
    )


jmrq_sandbox_text_value_pair = Struct("sandbox_text_value_pair",
    h3_string_id("parameter_name"),
    h3_reflexive("text_value_pairs", jmrq_sandbox_text_value_pair_text_value_pair),
    ENDIAN=">", SIZE=16
    )


jmrq_body = Struct("tagdata",
    h3_reflexive("sandbox_text_value_pairs", jmrq_sandbox_text_value_pair),
    ENDIAN=">", SIZE=12
    )


def get():
    return jmrq_def

jmrq_def = TagDef("jmrq",
    h3_blam_header('jmrq'),
    jmrq_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["jmrq"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

# DO NOT REMOVE! __all__ is used to locate frozen definitions
__all__ = (
    "ugh_", "ant_", "bitm", "trak", "snd_", "shad", "hsc_", "pphy",
    #"adlg", "mdlg", "bipd", "bsdt", "____", "devo", "whip", "char", "gldf",
    #"clwd", "coll", "coln", "colo", "cont", "bloc", "crea", "jpt_", "deca",
    #"DECR", "DECP", "dobc", "devi", "ctrl", "lifi", "mach", "udlg", "effe",
    #"eqip", "garb", "matg", "grhi", "hudg", "hmt_", "hud_", "item", "itmc",
    #"lens", "ligh", "MGS2", "tdtl", "foot", "mpdt", "metr", "hlmt", "jmad",
    #"mcsr", "unic", "mulg", "mply", "goof", "nhdt", "obje", "prt3", "PRTM",
    #"pmov", "fpch", "phys", "phmo", "pixl", "fog ", "proj", "mode", "sbsp",
    #"ltmp", "scnr", "ai__", "_ipd", "cin_", "clu_", "____", "_rea", "dec_",
    #"dc_s", "dgr_", "_qip", "_cen", "_sce", "sslt", "_igh", "trg_", "_ehi",
    #"_eap", "scen", "egor", "stem", "slit", "spas", "sky_", "sncl", "spk_",
    #"sfx_", "snde", "lsnd", "snmx", "ssce", "BooM", "styl", "sily", "unit",
    #"unhi", "wgtz", "skin", "wgit", "wigl", "vehi", "vehc", "vrtx", "weap",
    #"_fx_", "wphi", "weat", "wind"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...os_hek.defs.senv import *

dlm_comment = """DIRECTIONAL LIGHTMAP PROPERTIES
Special shader settings for when your map has directional lightmaps rendered for it."""

dlm_specular_map_comment = """DLM SPECULAR (COLOR) MAP
*RGB holds the specular color.
*ALPHA can be used as an exponent mask to provide finer control
over the size of the specular highlights."""

dlm_tint_comment = """DLM SPECULAR TINT AND BRIGHTNESS"""

dlm_specular_comment = """DLM SPECULAR INTENSITY
Exponent controls the highlight size, the bigger the exponent, the smaller the highlight.
Coefficient controls the brightness of the highlights."""

os_senv_ext = Struct("shader_environment_extension",
    Bool16("dlm_flags",
        "do_not_use_dlms",
        COMMENT=dlm_comment
        ),
    Pad(2),
    Float("bump_amount"),

    dependency_os("specular_color_map", "bitm", COMMENT=dlm_specular_map_comment),
    Float("specular_color_coefficient"),
    Float("specular_color_exponent"),
    Bool16("specular_flags",
        "alpha_as_exponent_mask",
        ),
    Pad(2),

    float_zero_to_one("perpendicular_brightness", COMMENT=dlm_tint_comment),
    QStruct("perpendicular_tint_color", INCLUDE=rgb_float),
    float_zero_to_one("parallel_brightness"),
    QStruct("parallel_tint_color", INCLUDE=rgb_float),

    Float("specular_lighting_exponent", COMMENT=dlm_specular_comment),
    Float("specular_lighting_coefficient"),

    SIZE=100,
    )

# replace the padding with an open sauce shader environment extension reflexive
senv_attrs = dict(senv_attrs)
senv_attrs[3] = reflexive("os_shader_environment_ext", os_senv_ext, 1)

senv_body = Struct("tagdata",
    shdr_attrs,
    senv_attrs,
    SIZE=836,
    )

def get():
    return senv_def

senv_def = TagDef("senv",
    blam_header('senv', 2),
    senv_body,

    ext=".shader_environment", endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .schi import *
from supyr_struct.defs.tag_def import TagDef
from .objs.shdr import ShdrTag

chicago_2_stage_maps = Struct("two_stage_map", INCLUDE=chicago_4_stage_maps)

scex_attrs = Struct("scex_attrs",
    # Shader Properties
    Struct("chicago_shader_extended",
        UInt8("numeric_counter_limit",
            MIN=0, MAX=255, SIDETIP="[0,255]"),  # [0,255]

        Bool8("chicago_shader_flags", *trans_shdr_properties),
        SEnum16("first_map_type", *trans_shdr_first_map_type),
        SEnum16("framebuffer_blend_function", *framebuffer_blend_functions),
        SEnum16("framebuffer_fade_mode", *render_fade_mode),
        SEnum16("framebuffer_fade_source", *function_outputs),
        Pad(2),
        ),

    #Lens Flare
    float_wu("lens_flare_spacing"),  # world units
    dependency("lens_flare", "lens"),
    reflexive("extra_layers", extra_layers_block, 4,
        DYN_NAME_PATH='.filepath'),
    reflexive("four_stage_maps", chicago_4_stage_maps, 4,
        DYN_NAME_PATH='.bitmap.filepath'),
    reflexive("two_stage_maps", chicago_2_stage_maps, 2,
        DYN_NAME_PATH='.bitmap.filepath'),
    Bool32("extra_flags",
        "dont_fade_active_camouflage",
        "numeric_countdown_timer"
        ),
    SIZE=80
    )

scex_body = Struct("tagdata",
    shdr_attrs,
    scex_attrs,
    SIZE=120
    )


def get():
    return scex_def

scex_def = TagDef("scex",
    blam_header('scex'),
    scex_body,

    ext=".shader_transparent_chicago_extended",
    endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.h2.common_descs import *
from reclaimer.meta.halo2_map import valid_h2_tags
from reclaimer.meta.halo1_map import tag_path_pointer, tag_index_array_pointer


h2_alpha_map_header = Struct("map header",
    UEnum32('head', ('head', 'head'), EDITABLE=False, DEFAULT='head'),
    UEnum32("version",
        ("halo2alpha", 7),
        ),
    UInt32("decomp len"),
    UInt32("unknown"),
    UInt32("tag index header offset"),
    UInt32("tag data size"),
    Pad(8),
    ascii_str32("map name"),
    ascii_str32("build date", EDITABLE=False),
    UEnum16("map type",
        "sp",
        "mp",
        "ui",
        "shared",
        ),
    Pad(2),
    UInt32("crc32"),
    UInt32("unknown0"),
    Pad(8),
    UInt32("unknown1"),  # seems to be a count?
    UInt32("crazy data offset"),
    UInt32("crazy data len"),
    UInt32("string id padded offset"),
    UInt32("string id count"),
    UInt32("string id table size"),
    UInt32("string id index offset"),
    UInt32("string id table offset"),
    BytesRaw("unknown_bytes", SIZE=16),

    Pad(1880),
    UEnum32('foot', ('foot', 'foot'), EDITABLE=False, DEFAULT='foot'),
    SIZE=2048
    )

tag_header = Struct("tag header",
    UEnum32("class 1", GUI_NAME="primary tag class", INCLUDE=valid_h2_tags),
    UEnum32("class 2", GUI_NAME="secondary tag class", INCLUDE=valid_h2_tags),
    UEnum32("class 3", GUI_NAME="tertiary tag class", INCLUDE=valid_h2_tags),
    UInt32("id"),
    UInt32("path offset"),
    UInt32("meta offset"),
    Bool32("flags", *("flag_%s" % i for i in range(32))),  # this is a guess
    UInt32("pad"),
    STEPTREE=CStrLatin1("path", POINTER=tag_path_pointer),
    SIZE=32
    )

tag_index_array = TagIndex("tag index",
    SIZE=".tag_count", SUB_STRUCT=tag_header, POINTER=tag_index_array_pointer
    )

h2_alpha_tag_index = Struct("tag index",
    UInt32("tag index offset"),
    UInt32("scenario tag id"),
    UInt32("unknown"),
    UInt32("tag count"),
    UInt32("tag sig", EDITABLE=False, DEFAULT='tags'),

    SIZE=20,
    STEPTREE=tag_index_array
    )

h2_alpha_map_header_def = BlockDef(h2_alpha_map_header)
h2_alpha_tag_index_def = BlockDef(h2_alpha_tag_index)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

fog_comment = """FLAGS
Setting <atmosphere dominant> prevents polygon popping when the atmospheric fog maximum
density (in the sky tag) is 1 and the atmospheric fog opaque distance is less than the
diameter of the map. However, this flag will cause artifacts when the camera goes below
the fog plane - so it should only be used when the fog plane is close to the ground."""

fog__body = Struct("tagdata",
    #fog flags
    Bool32("flags",
        "is_water",
        "atmospheric_dominant",
        "fog_screen_only",
        COMMENT=fog_comment
        ),

    Pad(84),
    #Density
    float_zero_to_one("maximum_density"),
    Pad(4),
    float_wu("opaque_distance"),
    Pad(4),
    float_wu("opaque_depth"),
    Pad(8),
    float_wu("distance_to_water_plane"),

    #Color
    QStruct("fog_color", INCLUDE=rgb_float),

    #Screen Layers
    Struct("screen_layers",
        Bool16("flags",
            "no_environment_multipass",
            "no_model_multipass",
            "no_texture_based_falloff",
            ),
        UInt16("layer_count", SIDETIP="[0,4]", MIN=0, MAX=4),

        from_to_wu("distance_gradient"),
        from_to_zero_to_one("density_gradient"),

        float_wu("start_distance_from_fog_plane"),
        Pad(4),

        #QStruct("color", INCLUDE=xrgb_byte),
        UInt32("color", INCLUDE=xrgb_uint32),
        float_zero_to_one("rotation_multiplier"),
        float_zero_to_one("strafing_multiplier"),
        float_zero_to_one("zoom_multiplier"),
        Pad(8),
        Float("map_scale"),
        dependency("fog_map", "bitm")
        ),

    #Screen Layer Animation
    Struct("screen_layer_animation",
        float_sec("animation_period"),
        Pad(4),
        from_to_wu_sec("wind_velocity"),
        from_to_sec("wind_period"),
        float_zero_to_one("wind_acceleration_weight"),
        float_zero_to_one("wind_perpendicular_weight")
        ),

    Pad(8),
    #Sound
    dependency("background_sound", "lsnd"),
    dependency("sound_environment", "snde"),
    SIZE=396,
    )

def get():
    return fog__def

fog__def = TagDef("fog ",
    blam_header('fog '),
    fog__body,

    ext=".fog", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .shpp import *
from ...hek.defs.objs.tag import HekTag

def merged_value_case(parent=None, **kwargs):
    if parent is None:
        raise ValueError()

    case  = parent.value_type.enum_name
    count = parent.value_count
    if count > 1:
        case += str(count)
    return case


value_name  = ascii_str32("value_name")
boolean_flags = Bool16("flags",
    "invert",
    )
runtime_value = SEnum16("runtime_value",
    "none",
    )
animation_flags1 = Bool8("animation_flags",
    "invert"
    )
animation_flags2 = Bool8("animation_flags",
    "invert",
    "multichannel_noise",
    )
animation_flags3 = Bool8("animation_flags",
    "invert",
    "multichannel_noise",
    "ignore_alpha",
    )
animation_function = SEnum16("animation_function", *animation_functions)

merged_value = Struct("merged_value",
    value_name,
    UEnum16("value_type",
        "boolean",
        "integer",
        "float",
        "color",
        "unused",
        "bitmap",
        ),
    UInt16("value_count", DEFAULT=1),
    Pad(4),
    Union("values",
        CASE=merged_value_case,
        CASES={
            "boolean": QStruct("value",
                UInt8("value", MIN=0, MAX=1, SIDETIP="[0 or 1]"),
                Pad(31)
                ),
            "integer": QStruct("value",
                SInt32("value_lower_bound"),
                SInt32("value_upper_bound"),
                Pad(24)
                ),
            "float": QStruct("value",
                Float("value_lower_bound"),
                Pad(12),
                Float("value_upper_bound"),
                Pad(12)
                ),
            "float2": Struct("value",
                QStruct("value_lower_bound", INCLUDE=ij_float),
                Pad(8),
                QStruct("value_upper_bound", INCLUDE=ij_float),
                Pad(8)
                ),
            "float3": Struct("value",
                QStruct("value_lower_bound", INCLUDE=ijk_float),
                Pad(4),
                QStruct("value_upper_bound", INCLUDE=ijk_float),
                Pad(4)
                ),
            "float4": Struct("value",
                QStruct("value_lower_bound", INCLUDE=ijkw_float),
                QStruct("value_upper_bound", INCLUDE=ijkw_float),
                ),
            "color": Struct("value",
                QStruct("value_lower_bound", INCLUDE=argb_float),
                QStruct("value_upper_bound", INCLUDE=argb_float),
                ),
            "bitmap": QStruct("value",
                SInt16("bitmap_index"),
                Pad(30),
                )
            },
        SIZE=32,
        ),

    # Runtime value override
    runtime_value,
    boolean_flags,

    # Animation
    animation_function,
    animation_flags3,
    Pad(1),

    float_sec("animation_duration"),
    Float("animation_rate", UNIT_SCALE=per_sec_unit_scale),

    dependency_os("bitmap", "bitm"),
    Pad(12),
    SIZE=116
    )

bitmap = Struct("bitmap",
    value_name,
    Pad(8),
    SInt16("bitmap_index"),
    Pad(30),

    Pad(28),
    dependency_os("bitmap", "bitm"),
    SIZE=116
    )

boolean = Struct("boolean",
    value_name,
    Pad(8),
    UInt8("value", MIN=0, MAX=1, SIDETIP="[0 or 1]"),
    Pad(31),

    runtime_value,
    boolean_flags,

    # Animation
    animation_function,
    animation_flags1,
    Pad(1),
    float_sec("animation_duration"),
    Float("animation_rate", UNIT_SCALE=per_sec_unit_scale),
    SIZE=88
    )

integer = Struct("integer",
    value_name,
    Pad(8),
    SInt32("value_lower_bound"),
    SInt32("value_upper_bound"),
    Pad(24),

    runtime_value,
    Pad(2),

    # Animation
    animation_function,
    animation_flags1,
    Pad(1),
    float_sec("animation_duration"),
    Float("animation_rate", UNIT_SCALE=per_sec_unit_scale),
    SIZE=88
    )

float_1d = Struct("float_1d",
    value_name,
    Pad(8),
    Float("value_lower_bound"),
    Pad(12),
    Float("value_upper_bound"),
    Pad(12),

    runtime_value,
    Pad(2),

    # Animation
    animation_function,
    animation_flags1,
    Pad(1),
    float_sec("animation_duration"),
    Float("animation_rate", UNIT_SCALE=per_sec_unit_scale),
    SIZE=88
    )

float_2d = Struct("float_2d",
    value_name,
    Pad(8),
    QStruct("value_lower_bound", INCLUDE=ij_float),
    Pad(8),
    QStruct("value_upper_bound", INCLUDE=ij_float),
    Pad(8),

    runtime_value,
    Pad(2),

    # Animation
    animation_function,
    animation_flags2,
    Pad(1),
    float_sec("animation_duration"),
    Float("animation_rate", UNIT_SCALE=per_sec_unit_scale),
    SIZE=88
    )

float_3d = Struct("float_3d",
    value_name,
    Pad(8),
    QStruct("value_lower_bound", INCLUDE=ijk_float),
    Pad(4),
    QStruct("value_upper_bound", INCLUDE=ijk_float),
    Pad(4),

    runtime_value,
    Pad(2),

    # Animation
    animation_function,
    animation_flags2,
    Pad(1),
    float_sec("animation_duration"),
    Float("animation_rate", UNIT_SCALE=per_sec_unit_scale),
    SIZE=88
    )

float_4d = Struct("float_4d",
    value_name,
    Pad(8),
    QStruct("value_lower_bound", INCLUDE=ijkw_float),
    QStruct("value_upper_bound", INCLUDE=ijkw_float),

    runtime_value,
    Pad(2),

    # Animation
    animation_function,
    animation_flags2,
    Pad(1),
    float_sec("animation_duration"),
    Float("animation_rate", UNIT_SCALE=per_sec_unit_scale),
    SIZE=88
    )

color = Struct("color",
    value_name,
    Pad(8),
    QStruct("value_lower_bound", INCLUDE=argb_float),
    QStruct("value_upper_bound", INCLUDE=argb_float),

    # Runtime value override
    SEnum16("runtime_value",
        "none",
        "player_team_color",
        ),
    Pad(2),

    # Animation
    animation_function,
    animation_flags3,
    Pad(1),
    float_sec("animation_duration"),
    Float("animation_rate", UNIT_SCALE=per_sec_unit_scale),
    SIZE=88
    )

shpg_attrs = Struct("shpg_attrs",
    Pad(4),
    dependency_os("base_shader", "shpg"),

    # when compiled into a map, all values are merged into this array
    reflexive("merged_values", merged_value, 16*8,
              DYN_NAME_PATH='.value_name', VISIBLE=False),
    reflexive("bitmaps",   bitmap,   16, DYN_NAME_PATH='.value_name'),
    reflexive("booleans",  boolean,  16, DYN_NAME_PATH='.value_name'),
    reflexive("integers",  integer,  16, DYN_NAME_PATH='.value_name'),
    reflexive("floats_1d", float_1d, 16, DYN_NAME_PATH='.value_name'),
    reflexive("floats_2d", float_2d, 16, DYN_NAME_PATH='.value_name'),
    reflexive("floats_3d", float_3d, 16, DYN_NAME_PATH='.value_name'),
    reflexive("floats_4d", float_4d, 16, DYN_NAME_PATH='.value_name'),
    reflexive("colors",    color,    16, DYN_NAME_PATH='.value_name'),
    SIZE=164
    )

shpg_body = Struct("tagdata",
    shpp_attrs,
    shpg_attrs,
    SIZE=328
    )

def get():
    return shpg_def

shpg_def = TagDef("shpg",
    blam_header_os('shpg'),
    shpg_body,

    ext=".shader_postprocess_generic", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Basically done.
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


wspr_animation_definition = Struct("animation_definition",
    UInt32("frame"),
    SInt16("sprite_index_0"),
    SInt16("sprite_index_1"),
    BytesRaw("unknown", SIZE=12, VISIBLE=False),
    ENDIAN=">", SIZE=20
    )


wspr_body = Struct("tagdata",
    Bool32("animation_flags",
        "loops",
        ),
    h3_reflexive("animation_definition", wspr_animation_definition),
    h3_rawdata_ref("data"),
    ENDIAN=">", SIZE=36
    )


def get():
    return wspr_def

wspr_def = TagDef("wspr",
    h3_blam_header('wspr'),
    wspr_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["wspr"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.common_descs import *

from supyr_struct.defs.bitmaps.dds import pixelformat_typecodes
from supyr_struct.defs.block_def import BlockDef


d3d_texture = Struct("d3d_texture",
    UInt16("width"),
    UInt16("height"),
    UInt8("depth"),
    UInt8("texture_count"), # mipmap_count + 1
    UEnum8("type",
        "texture_2d",
        "texture_3d",
        "cubemap",
        "lightmap",
        ),
    UInt8("tex_page_index"),  # page the fullsize image is stored in.
    #                           mipmaps are always stored in primary.
    UEnum32("format", *pixelformat_typecodes)
    )


s_tag_d3d_texture = Struct("s_tag_d3d_texture",
    Struct("primary_page_data", INCLUDE=rawdata_ref_struct),
    Struct("secondary_page_data", INCLUDE=rawdata_ref_struct),
    Struct("texture", INCLUDE=d3d_texture)
    )


s_tag_d3d_texture_interleaved = Struct("s_tag_d3d_texture_interleaved",
    Struct("primary_page_data", INCLUDE=rawdata_ref_struct),
    Struct("secondary_page_data", INCLUDE=rawdata_ref_struct),
    Struct("texture0", INCLUDE=d3d_texture),
    Struct("texture1", INCLUDE=d3d_texture),
    )

s_tag_d3d_texture_def = BlockDef(
    s_tag_d3d_texture,
    endian=">"
    )
s_tag_d3d_texture_interleaved_def = BlockDef(
    s_tag_d3d_texture_interleaved,
    endian=">"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.scen import *
from .obje import *

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=6)

scen_body = dict(scen_body)
scen_body[0] = obje_attrs

def get():
    return scen_def

scen_def = TagDef("scen",
    blam_header('scen'),
    scen_body,

    ext=".scenery", endian=">", tag_cls=ObjeTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from array import array
from reclaimer.h2.common_descs import *
from supyr_struct.defs.tag_def import TagDef

def pixel_block_size(node, *a, **kwa):
    if isinstance(node, array):
        return node.itemsize*len(node)
    return len(node)

pixel_root = WhileArray('pixel root',
    SUB_STRUCT=WhileArray('bitmap pixels',
        SUB_STRUCT=UInt8Array('pixels', SIZE=pixel_block_size)
        )
    )

sprite = QStruct("sprite",
    UInt16("bitmap index"),
    Pad(6),
    Float("left side"),
    Float("right side"),
    Float("top side"),
    Float("bottom side"),
    Float("registration point x"),
    Float("registration point y"),
    SIZE=32
    )

sequence = Struct("sequence",
    ascii_str32("sequence name"),
    UInt16("first bitmap index"),
    UInt16("bitmap count"),
    Pad(16),
    h2_reflexive("sprites", sprite, 64),
    SIZE=60
    )

bitmap = Struct("bitmap",
    UEnum32('bitm id', ('bitm', 'bitm'), DEFAULT='bitm', EDITABLE=False),
    UInt16("width", SIDETIP="pixels", EDITABLE=False),
    UInt16("height", SIDETIP="pixels", EDITABLE=False),
    UInt8("depth", SIDETIP="pixels", EDITABLE=False),
    Bool8("more flags",
        "delete from cache file",
        "bitmap create attempted",
        "unknown"
        ),
    SEnum16("type",
        "texture 2d",
        "texture 3d",
        "cubemap",
        "white",
        EDITABLE=False
        ),
    SEnum16("format",
        "a8",
        "y8",
        "ay8",
        "a8y8",
        #"-unused1-",
        #"-unused2-",
        ("r5g6b5", 6),
        #"-unused3-",
        ("a1r5g5b5", 8),
        ("a4r4g4b4", 9),
        ("x8r8g8b8", 10),
        ("a8r8g8b8", 11),
        #"-unused4-",
        #"-unused5-",
        ("dxt1", 14),
        ("dxt3", 15),
        ("dxt5", 16),
        ("p8 bump", 17),
        ("p8", 18),
        ("argbfp32", 19),
        ("rgbfp32", 20),
        ("rgbfp16", 21),
        ("v8u8", 22),
        ("g8b8", 23),
        ),
    Bool16("flags",
        "power of 2 dim",
        "compressed",
        "palletized",
        "swizzled",
        "linear",
        "v16u16",
        "mipmap debug level",
        "prefer low detail",
        ),
    UInt16("registration point x"),
    UInt16("registration point y"),
    UInt16("mipmaps"),
    UInt16("low detail mipmaps"),

    UInt32("pixels offset", EDITABLE=False),
    UInt32("lod1 offset", EDITABLE=False),
    UInt32("lod2 offset", EDITABLE=False),
    UInt32("lod3 offset", EDITABLE=False),
    UInt32("lod4 offset", EDITABLE=False),
    UInt32("lod5 offset", EDITABLE=False),
    UInt32("lod6 offset", EDITABLE=False),
    UInt32("lod1 size", EDITABLE=False),
    UInt32("lod2 size", EDITABLE=False),
    UInt32("lod3 size", EDITABLE=False),
    UInt32("lod4 size", EDITABLE=False),
    UInt32("lod5 size", EDITABLE=False),
    UInt32("lod6 size", EDITABLE=False),
    UInt32("datum",  # points back to this tag
        VISIBLE=False, EDITABLE=False
        ),
    SIZE=116
    )

bitm_body = Struct("tagdata",
    SEnum16("type",
        "textures 2d",
        "textures 3d",
        "cubemaps",
        "sprites",
        "interface bitmaps",
        ),
    SEnum16("format",
        "color key transparency",
        "explicit alpha",
        "interpolated alpha",
        "color 16bit",
        "color 32bit",
        "monochrome",
        ),
    SEnum16("usage",
        "alpha-blend",
        "default",
        "height map",
        "detail map",
        "light map",
        "vector map",
        "height map blue 255",
        "embm",
        "height map a8l8",
        "height map g8b8",
        "height map g8b8 with alpha",
        DEFAULT=1
        ),
    Bool16("flags",
        "enable diffusion dithering",
        "disable height map compression",
        "uniform sprite sequences",
        "sprite bug fix",
        "use sharp bump filter",
        "unused",
        "use clamped bump filter",
        "invert detail fade",
        "swap x y vector components",
        "convert from signed",
        "convert to signed",
        "import mipmap chains",
        "intentionally true color"
        ),
    QStruct("post processing",
        float_zero_to_one("detail fade factor"),
        float_zero_to_one("sharpen amount"),
        Float("bump height", SIDETIP="repeats"),
        ),
    Struct("sprite processing",
        SEnum16("sprite budget size",
          {NAME: "x32",  VALUE: 0, GUI_NAME: "32x32"},
          {NAME: "x64",  VALUE: 1, GUI_NAME: "64x64"},
          {NAME: "x128", VALUE: 2, GUI_NAME: "128x128"},
          {NAME: "x256", VALUE: 3, GUI_NAME: "256x256"},
          {NAME: "x512", VALUE: 4, GUI_NAME: "512x512"},
          {NAME: "x1024", VALUE: 5, GUI_NAME: "1024x1024"},
          ),
        UInt16("sprite budget count"),
        ),
    UInt16("color plate width",  SIDETIP="pixels", EDITABLE=False),
    UInt16("color plate height", SIDETIP="pixels", EDITABLE=False),

    h2_rawdata_ref("compressed color plate data", max_size=16777216),
    h2_rawdata_ref("processed pixel data", max_size=16777216),

    Float("blur filter size", MIN=0.0, MAX=10.0, SIDETIP="[0,10] pixels"),
    float_neg_one_to_one("alpha bias"),
    UInt16("mipmap levels", MIN=0, SIDETIP="levels"),
    SEnum16("sprite usage",
       "blend/add/subtract/max",
       "multiply/min",
       "double multiply",
       ),
    UInt16("sprite spacing", SIDETIP="pixels"),
    SEnum16("force format",
       "default",
       "g8b8",
       "dxt1",
       "dxt3",
       "dxt5",
       "a8l8",
       "a4r4g4b4",
       ),
    h2_reflexive("sequences", sequence, 256, DYN_NAME_PATH='.sequence_name'),
    h2_reflexive("bitmaps", bitmap, 65536),
    Struct("more sprite processing",
        SInt8("color compression quality", SIDETIP="[1, 127]"),
        SInt8("alpha compression quality", SIDETIP="[1, 127]"),
        SInt8("overlap"),
        UEnum8("color subsampling",
            {GUI_NAME:"4:0:0", NAME:"x0x0"},
            {GUI_NAME:"4:2:0", NAME:"x2x0"},
            {GUI_NAME:"4:2:2", NAME:"x2x2"},
            {GUI_NAME:"4:4:4", NAME:"x4x4"},
            )
        ),
    ENDIAN="<", SIZE=76, WIDGET=Halo2BitmapTagFrame,
    )


def get():
    return bitm_def

bitm_def = TagDef("bitm",
    h2_blam_header('bitm'),
    bitm_body,

    ext=".%s" % h2_tag_class_fcc_to_ext["bitm"], endian="<",
    subdefs={'pixel_root':pixel_root}
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

'''
    This variant of the gametype tag is intended to allow easily creating a
    gametype from within Mozzarilla by just selecting the xbox_gametype def
'''

from reclaimer.misc.defs.pc_gametype import xbox_gametype_header, settings,\
     xbox_gametype_footer
from reclaimer.misc.defs.objs.xbox_gametype import XboxGametypeTag

from supyr_struct.defs.tag_def import TagDef

def get():
    return xbox_gametype_def

xbox_gametype_def = TagDef('xbox_gametype',
    xbox_gametype_header,
    settings,
    xbox_gametype_footer,

    ext='.lst', endian='<', tag_cls=XboxGametypeTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

sidy_body = Struct("tagdata",
    rawdata_ref("documentation", StrAscii, max_size=65536),
    SIZE=40
    )

def get():
    return sidy_def

sidy_def = TagDef("sidy",
    blam_header_os('sidy'),
    sidy_body,

    ext=".string_id_yelo", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

change_color = Struct("change_color",
    QStruct("color_lower_bound", INCLUDE=rgb_float),
    QStruct("color_upper_bound", INCLUDE=rgb_float),
    SIZE=32
    )

actv_body = Struct("tagdata",
    Bool32('flags',
        "can_shoot_while_flying",
        "blend_color_in_hsv",
        "has_unlimited_grenades",
        "moveswitch_stay_with_friends",
        "active_camouflage",
        "super_active_camouflage",
        "cannot_use_ranged_weapons",
        "prefer_passenger_seat",
        ),
    dependency("actor_definition", "actr"),
    dependency("unit", valid_units),
    dependency("major_variant", "actv"),

    #Movement switching
    Struct("movement_switching",
        Pad(24),
        SEnum16("movement_type",
            "always_run",
            "always_crouch",
            "switch_types",
            ),
        Pad(2),
        float_zero_to_one("initial_crouch_chance"),
        from_to_sec("crouch_time", UNIT_SCALE=sec_unit_scale),  # seconds
        from_to_sec("run_time",    UNIT_SCALE=sec_unit_scale)   # seconds
        ),

    #Ranged combat
    Struct("ranged_combat",
        dependency("weapon", "weap"),
        float_wu("maximum_firing_distance"),
        Float("rate_of_fire", UNIT_SCALE=per_sec_unit_scale),  # rounds/sec
        float_rad("projectile_error"),  # radians
        from_to_sec("first_burst_delay_time",
                    UNIT_SCALE=sec_unit_scale),  # seconds
        Float("new_target_firing_pattern_time",
               UNIT_SCALE=sec_unit_scale),  # seconds
        Float("surprise_delay_time", UNIT_SCALE=sec_unit_scale),  # seconds
        Float("surprise_fire_wildly_time",
               UNIT_SCALE=sec_unit_scale),  # seconds
        float_zero_to_one("death_fire_wildly_chance"),
        float_sec("death_fire_wildly_time",
                  UNIT_SCALE=sec_unit_scale),  # seconds
        from_to_wu("desired_combat_range"),
        QStruct("custom_stand_gun_offset", INCLUDE=ijk_float),
        QStruct("custom_crouch_gun_offset", INCLUDE=ijk_float),
        float_zero_to_one("target_tracking"),
        float_zero_to_one("target_leading"),
        Float("weapon_damage_modifier"),
        Float("damage_per_second", UNIT_SCALE=per_sec_unit_scale),  # seconds
        ),

    #Burst geometry
    Struct("burst_geometry",
        float_wu("burst_origin_radius"),
        float_rad("burst_origin_angle"),  # radians
        from_to_sec("burst_return_length"),
        float_rad("burst_return_angle"),  # radians
        from_to_sec("burst_duration", UNIT_SCALE=sec_unit_scale),  # seconds
        from_to_sec("burst_separation"),
        Float("burst_angular_velocity", SIDETIP="degrees/sec",
               UNIT_SCALE=irad_per_sec_unit_scale), # radians/second
        Pad(4),
        float_zero_to_one("special_damage_modifier"),
        float_rad("special_projectile_error")  # radians
        ),

    #Firing patterns"
    Struct("firing_patterns",
        Float("new_target_burst_duration"),
        Float("new_target_burst_separation"),
        Float("new_target_rate_of_fire", UNIT_SCALE=per_sec_unit_scale),
        Float("new_target_projectile_error"),

        Pad(8),
        Float("moving_burst_duration"),
        Float("moving_burst_separation"),
        Float("moving_rate_of_fire", UNIT_SCALE=per_sec_unit_scale),
        Float("moving_projectile_error"),

        Pad(8),
        Float("berserk_burst_duration"),
        Float("berserk_burst_separation"),
        Float("berserk_rate_of_fire", UNIT_SCALE=per_sec_unit_scale),
        Float("berserk_projectile_error")
        ),

    #Special-case firing patterns
    Struct("special_case_firing_patterns",
        Pad(8),
        Float("super_ballistic_range"),
        Float("bombardment_range"),
        Float("modified_vision_range"),
        SEnum16("special_fire_mode",
            "none",
            "overcharge",
            "secondary_trigger",
            ),
        SEnum16("special_fire_situation",
            "never",
            "enemy_visible",
            "enemy_out_of_sight",
            "strafing",
            ),
        float_zero_to_one("special_fire_chance"),
        float_sec("special_fire_delay", UNIT_SCALE=sec_unit_scale)
        ),

    #Berserking and melee
    Struct("berserking_and_melee",
        float_wu("melee_range"),
        float_wu("melee_abort_range"),
        from_to_wu("berserk_firing_ranges", INCLUDE=from_to),
        float_wu("berserk_melee_range"),
        float_wu("berserk_melee_abort_range")
        ),

    #Grenades
    Struct("grenades",
        Pad(8),
        SEnum16("grenade_type", *grenade_types),
        SEnum16("trajectory_type",
            "toss",
            "lob",
            "bounce",
            ),
        SEnum16("grenade_stimulus",
            "never",
            "visible_target",
            "seek_cover",
            ),
        SInt16("minimum_enemy_count"),
        float_wu("enemy_radius"),

        Pad(4),
        float_wu_sec("grenade_velocity", UNIT_SCALE=per_sec_unit_scale),
        from_to_wu("grenade_ranges"),
        float_wu("collateral_damage_radius"),
        float_zero_to_one("grenade_chance"),
        float_sec("grenade_check_time", UNIT_SCALE=sec_unit_scale),
        float_sec("encounter_grenade_timeout", UNIT_SCALE=sec_unit_scale)
        ),

    #Items
    Struct("items",
        Pad(20),
        dependency("equipment", "eqip"),
        QStruct("grenade_count",
            SInt16("from", GUI_NAME=""), SInt16("to"), ORIENT='h'
            ),
        float_zero_to_one("dont_drop_grenades_chance"),
        QStruct("drop_weapon_loaded", INCLUDE=from_to),
        QStruct("drop_weapon_ammo",
            SInt16("from", GUI_NAME=""),
            SInt16("to"), ORIENT='h'
            )
        ),

    #Unit properties
    Struct("unit_properties",
        Pad(28),
        Float("body_vitality"),
        Float("shield_vitality"),
        float_wu("shield_sapping_radius"),
        SInt16("forced_shader_permutation"),
        ),

    Pad(30),
    reflexive("change_colors", change_color, 4),
    SIZE=568
    )


def get():
    return actv_def

actv_def = TagDef("actv",
    blam_header('actv'),
    actv_body,

    ext=".actor_variant", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Portin'
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

proj_detonation_timer_starts = (
    "immediately",
    "after_first_bounce",
    "when_at_rest",
    "after_first_bounce_off_any_surface",
    )

proj_material_response_response_0 = (
    "impact_detonate",
    "fizzle",
    "overpenetrate",
    "attach",
    "bounce",
    "bounce_dud",
    "fizzle_ricochet",
    )

proj_material_response_scale_effects_by = (
    "damage",
    "angle",
    )


proj_early_mover_propertie = Struct("early_mover_propertie",
    h3_string_id("name"),
    Pad(36),
    ENDIAN=">", SIZE=40
    )


proj_ai_propertie = Struct("ai_propertie",
    Bool32("flags",
        "destroyable_cover",
        "pathfinding_ignore_when_dead",
        "dynamic_cover",
        ),
    h3_string_id("ai_type_name"),
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    SEnum16("size", *bloc_ai_propertie_size),
    SEnum16("leap_jump_speed", *bloc_ai_propertie_leap_jump_speed),
    ENDIAN=">", SIZE=16
    )


proj_function = Struct("function",
    Bool32("flags",
        "invert",
        "mapping_does_not_controls_active",
        "always_active",
        "random_time_offset",
        ),
    h3_string_id("import_name"),
    h3_string_id("export_name"),
    h3_string_id("turn_off_with"),
    Float("minimum_value"),
    h3_rawdata_ref("default_function"),
    h3_string_id("scale_by"),
    ENDIAN=">", SIZE=44
    )


proj_attachment = Struct("attachment",
    h3_dependency("attachment"),
    h3_string_id("marker"),
    SEnum16("change_color", *bloc_attachment_change_color),
    SInt16("unknown", VISIBLE=False),
    h3_string_id("primary_scale"),
    h3_string_id("secondary_scale"),
    ENDIAN=">", SIZE=32
    )


proj_widget = Struct("widget",
    h3_dependency("type"),
    ENDIAN=">", SIZE=16
    )


proj_change_color_initial_permutation = Struct("initial_permutation",
    Pad(4),
    color_rgb_float("color_lower_bound"),
    color_rgb_float("color_upper_bound"),
    h3_string_id("variant_name"),
    ENDIAN=">", SIZE=32
    )


proj_change_color_function = Struct("function",
    Bool32("scale_flags",
        "blend_in_hsv",
        "more_colors",
        ),
    color_rgb_float("color_lower_bound"),
    color_rgb_float("color_upper_bound"),
    h3_string_id("darken_by"),
    h3_string_id("scale_by"),
    ENDIAN=">", SIZE=32
    )


proj_change_color = Struct("change_color",
    h3_reflexive("initial_permutations", proj_change_color_initial_permutation),
    h3_reflexive("functions", proj_change_color_function),
    ENDIAN=">", SIZE=24
    )


proj_predicted_resource = Struct("predicted_resource",
    SInt16("type"),
    SInt16("resource_index"),
    dependency_uint32("tag_index", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=8
    )


proj_multiplayer_object_propertie = Struct("multiplayer_object_propertie",
    Bool16("engine_flags",
        "capture_the_flag",
        "slayer",
        "oddball",
        "king_of_the_hill",
        "juggernaut",
        "territories",
        "assault",
        "vip",
        "infection",
        ),
    SEnum8("object_type", *bloc_multiplayer_object_propertie_object_type),
    Bool8("teleporter_flags",
        "disallows_players",
        "allows_land_vehicles",
        "allows_heavy_vehicles",
        "allows_flying_vehicles",
        "allows_projectiles",
        ),
    Bool16("flags",
        "editor_only",
        ),
    SEnum8("shape", *bloc_multiplayer_object_propertie_shape),
    SEnum8("spawn_timer_mode", *bloc_multiplayer_object_propertie_spawn_timer_mode),
    SInt16("spawn_time"),
    SInt16("abandon_time"),
    Float("radius_width"),
    Float("length"),
    Float("top"),
    Float("bottom"),
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    SInt32("unknown_3", VISIBLE=False),
    SInt32("unknown_4", VISIBLE=False),
    h3_dependency("child_object"),
    SInt32("unknown_5"),
    h3_dependency("shape_shader"),
    h3_dependency("unknown_shader"),
    h3_dependency("unknown_6"),
    h3_dependency("unknown_7"),
    h3_dependency("unknown_8"),
    h3_dependency("unknown_9"),
    h3_dependency("unknown_10"),
    h3_dependency("unknown_11"),
    ENDIAN=">", SIZE=196
    )


proj_material_response = Struct("material_response",
    Bool16("flags_0",
        "cannot_be_overpenetrated",
        ),
    SEnum16("response_0", *proj_material_response_response_0),
    h3_string_id("material_name"),
    SInt16("global_material_index"),
    SInt16("unknown_0", VISIBLE=False),
    SEnum16("response_1", *proj_material_response_response_0),
    Bool16("flags_1",
        "only_against_units",
        "never_against_units",
        ),
    Float("chance_fraction"),
    from_to_rad("between_angle"),
    QStruct("and_velocity", INCLUDE=from_to),
    SEnum16("scale_effects_by", *proj_material_response_scale_effects_by),
    SInt16("unknown_1", VISIBLE=False),
    float_rad("angular_noise"),
    Float("velocity_noise"),
    Float("initial_friction"),
    Float("maximum_distance"),
    Float("parallel_friction"),
    Float("perpendicular_friction"),
    ENDIAN=">", SIZE=64
    )


proj_claymore_grenade = Struct("claymore_grenade",
    float_rad("unknown_0"),
    float_rad("unknown_1"),
    float_rad("unknown_2"),
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=9),
    ENDIAN=">", SIZE=48
    )


proj_firebomb_grenade = Struct("firebomb_grenade",
    Float("unknown"),
    ENDIAN=">", SIZE=4
    )


proj_shotgun_propertie = Struct("shotgun_propertie",
    SInt16("amount"),
    SInt16("distance"),
    Float("accuracy"),
    float_rad("spread_cone_angle"),
    ENDIAN=">", SIZE=12
    )


proj_body = Struct("tagdata",
    SEnum16("object_type", *bloc_object_type),
    Bool16("flags_0",
        "does_not_cast_shadow",
        "search_cardinal_direction_lightmaps",
        ("not_a_pathfinding_obstacle", 1 << 3),
        "extension_of_parent",
        "does_not_cause_collision_damage",
        "early_mover",
        "early_mover_localized_physics",
        "use_static_massive_lightmap_sample",
        "object_scales_attachments",
        "inherits_player_s_appearance",
        "dead_bipeds_can_t_localize",
        "attach_to_clusters_by_dynamic_sphere",
        "effects_created_by_this_object_do_not_spawn_objects_in_multiplayer",
        ),
    Float("bounding_radius"),
    QStruct("bounding_offset", INCLUDE=xyz_float),
    Float("acceleration_scale"),
    SEnum16("lightmap_shadow_mode_size", *bloc_lightmap_shadow_mode_size),
    SEnum8("sweetener_size", *bloc_sweetener_size),
    SEnum8("water_density", *bloc_water_density),
    SInt32("unknown_0", VISIBLE=False),
    Float("dynamic_light_sphere_radius"),
    QStruct("dynamic_light_sphere_offset", INCLUDE=xyz_float),
    h3_string_id("default_model_variant"),
    h3_dependency("model"),
    h3_dependency("crate_object"),
    h3_dependency("collision_damage"),
    h3_reflexive("early_mover_properties", proj_early_mover_propertie),
    h3_dependency("creation_effect"),
    h3_dependency("material_effects"),
    h3_dependency("melee_impact"),
    h3_reflexive("ai_properties", proj_ai_propertie),
    h3_reflexive("functions", proj_function),
    SInt16("hud_text_message_index"),
    SInt16("unknown_1", VISIBLE=False),
    h3_reflexive("attachments", proj_attachment),
    h3_reflexive("widgets", proj_widget),
    h3_reflexive("change_colors", proj_change_color),
    h3_reflexive("predicted_resources", proj_predicted_resource),
    h3_reflexive("multiplayer_object_properties", proj_multiplayer_object_propertie),
    Bool32("flags_1",
        "oriented_along_velocity",
        "ai_must_use_ballistic_aiming",
        "detonation_max_time_if_attached",
        "has_super_combining_explosion",
        "damage_scales_based_on_distance",
        "travels_instantaneously",
        "steering_adjusts_orientation",
        "dont_noise_up_steering",
        "can_track_behind_itself",
        "robotron_steering",
        "faster_when_owned_by_player",
        ),
    SEnum16("detonation_timer_starts", *proj_detonation_timer_starts),
    SEnum16("impact_noise", *gint_constant_sound_volume),
    Float("collision_radius"),
    Float("arming_time"),
    Float("danger_radius"),
    QStruct("timer", INCLUDE=from_to),
    Float("minimum_velocity"),
    Float("maximum_range"),
    Float("detonation_charge_time"),
    SEnum16("detonation_noise", *gint_constant_sound_volume),
    SInt16("super_detonation_projectile_count"),
    Float("super_detonation_delay"),
    h3_dependency("detonation_started"),
    h3_dependency("airborne_detonation_effect"),
    h3_dependency("ground_detonation_effect"),
    h3_dependency("detonation_damage"),
    h3_dependency("attached_detonation_damage"),
    h3_dependency("super_detonation"),
    h3_dependency("super_detonation_damage"),
    h3_dependency("detonation_sound"),
    SEnum8("damage_reporting_type", *proj_damage_reporting_type),
    SInt8("unknown_2", VISIBLE=False),
    SInt8("unknown_3", VISIBLE=False),
    SInt8("unknown_4", VISIBLE=False),
    h3_dependency("attached_super_detonation_damage"),
    Float("material_effect_radius"),
    h3_dependency("flyby_sound"),
    h3_dependency("flyby_response"),
    h3_dependency("impact_effect"),
    h3_dependency("impact_damage"),
    Float("boarding_detonation_time"),
    h3_dependency("boarding_detonation_damage"),
    h3_dependency("boarding_attached_detonation_damage"),
    Float("air_gravity_scale"),
    QStruct("air_damage_range", INCLUDE=from_to),
    Float("water_gravity_scale"),
    QStruct("water_damage_scale", INCLUDE=from_to),
    Float("initial_velocity"),
    Float("final_velocity"),
    Float("unknown_5", VISIBLE=False),
    Float("unknown_6", VISIBLE=False),
    float_rad("guided_angular_velocity_lower"),
    float_rad("guided_angular_velocity_upper"),
    float_rad("unknown_7"),
    QStruct("acceleration_range", INCLUDE=from_to),
    Float("unknown_8", VISIBLE=False),
    BytesRaw("unknown_9", SIZE=4, VISIBLE=False),
    Float("targeted_leading_fraction"),
    BytesRaw("unknown_10", SIZE=4, VISIBLE=False),
    h3_reflexive("material_responses", proj_material_response),
    h3_reflexive("claymore_grenade", proj_claymore_grenade),
    h3_reflexive("firebomb_grenade", proj_firebomb_grenade),
    h3_reflexive("shotgun_properties", proj_shotgun_propertie),
    ENDIAN=">", SIZE=672
    )


def get():
    return proj_def

proj_def = TagDef("proj",
    h3_blam_header('proj'),
    proj_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["proj"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

shader_function = Struct('shader_function',
    Pad(4),
    ascii_str32("global_function_name"),
    SIZE=36
    )

animation = Struct('animation',
    SInt16('animation_index'),
    Pad(2),
    float_sec("period"),
    SIZE=36
    )

light = Struct('light',
    dependency("lens_flare", "lens"),
    ascii_str32("global_function_name"),
    Pad(28),
    Bool32('flags',
        'affects_exteriors',
        'affects_interiors',
        ),
    QStruct("color", INCLUDE=rgb_float),
    Float("power"),
    Float("test_distance"),
    Pad(4),
    yp_float_rad("direction"),  # radians
    float_rad("diameter"),  # radians (yeah, it sounds weird, but this
    #                         value is stored as a radian coefficient)
    SIZE=116
    )


sky__body = Struct("tagdata",
    dependency("model", valid_models),
    dependency("animation_graph", "antr"),
    Pad(24),

    QStruct("indoor_ambient_radiosity_color", INCLUDE=rgb_float),
    Float("indoor_ambient_radiosity_power"),

    QStruct("outdoor_ambient_radiosity_color", INCLUDE=rgb_float),
    Float("outdoor_ambient_radiosity_power"),

    QStruct("outdoor_fog_color", INCLUDE=rgb_float),
    Pad(8),
    float_zero_to_one("outdoor_fog_maximum_density"),
    float_wu("outdoor_fog_start_distance"),
    float_wu("outdoor_fog_opaque_distance"),

    QStruct("indoor_fog_color", INCLUDE=rgb_float),
    Pad(8),
    float_zero_to_one("indoor_fog_maximum_density"),
    float_wu("indoor_fog_start_distance"),
    float_wu("indoor_fog_opaque_distance"),

    dependency("indoor_fog_screen", "fog "),
    Pad(4),
    reflexive("shader_functions", shader_function, 8,
        DYN_NAME_PATH='.global_function_name'),
    reflexive("animations", animation, 8),
    reflexive("lights", light, 8,
        DYN_NAME_PATH='.lens_flare.filepath'),

    SIZE=208,
    )


def get():
    return sky__def

sky__def = TagDef("sky ",
    blam_header('sky '),
    sky__body,

    ext=".sky", endian=">", tag_cls=HekTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.h2.common_descs import *
from supyr_struct.defs.tag_def import TagDef

pphy_body = BlockDef("tagdata",
    Bool32("flags",
        "flamethrower particle collision",
        "collides with structures",
        "collides with water surface",
        "uses simple wind",
        "uses damped wind",
        "no gravity"
        ),
    #these next three are courtesy of Sparky. I had
    #no idea these existed till I looked in Eschaton
    Float("wind coefficient", VISIBLE=False),
    Float("wind sine modifier", VISIBLE=False),
    Float("z translation rate", VISIBLE=False),
    Pad(16),
    Float("density", SIDETIP="g/mL"),#g/mL
    Float("air friction"),
    Float("water friction"),
    Float("surface friction"),
    Float("elasticity"),
    ENDIAN="<", SIZE=64
    )


def get():
    return pphy_def

pphy_def = TagDef("pphy",
    h2_blam_header('pphy'),
    pphy_body,

    ext=".%s" % h2_tag_class_fcc_to_ext["pphy"], endian="<"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

particle_type = Struct("particle_type",
    ascii_str32("name"),
    Bool32("flags",
        "blend_colors_in_hsv",
        "along_long_hue_path",
        "random_rotation",
        ),
    QStruct("fade",
        float_wu("in_start_distance"),
        float_wu("in_end_distance"),
        float_wu("out_start_distance"),
        float_wu("out_end_distance"),
        float_wu("in_start_height"),
        float_wu("in_end_height"),
        float_wu("out_start_height"),
        float_wu("out_end_height"),
        ),

    Pad(96),
    QStruct("particle_count", INCLUDE=from_to,
        SIDETIP="particles/(world unit^3)"),
    dependency("physics", "pphy"),

    Pad(16),
    Struct("acceleration",
        QStruct("magnitude", INCLUDE=from_to),
        float_rad("turning_rate"),  # radians
        Float("change_rate"),
        ),

    Pad(32),
    from_to_wu("particle_radius"),
    QStruct("animation_rate",
        Float("from", UNIT_SCALE=per_sec_unit_scale, GUI_NAME=''),
        Float("to",   UNIT_SCALE=per_sec_unit_scale),
        ORIENT='h', SIDETIP="frames/sec"
        ),
    from_to_rad_sec("rotation_rate"),  # radians/sec

    Pad(32),
    QStruct("color_lower_bound", INCLUDE=argb_float),
    QStruct("color_upper_bound", INCLUDE=argb_float),

    #Shader
    Struct("shader",
        Pad(64),
        dependency("sprite_bitmap", "bitm"),
        SEnum16("render_mode", *render_mode),
        SEnum16("render_direction_source",
            "from_velocity",
            "from_acceleration"
            ),

        Pad(40),
        Bool16("shader_flags", *shader_flags),
        SEnum16("framebuffer_blend_function", *framebuffer_blend_functions),
        SEnum16("framebuffer_fade_mode", *render_fade_mode),
        Bool16("map_flags",
            "unfiltered"
            )
        ),

    #Secondary bitmap
    Struct("secondary_bitmap",
        Pad(28),
        dependency("bitmap", "bitm"),
        SEnum16("anchor", *render_anchor),
        Bool16("secondary_map_flags",
            "unfiltered"
            ),
        Struct("u_animation", INCLUDE=anim_src_func_per_pha_sca),
        Struct("v_animation", INCLUDE=anim_src_func_per_pha_sca),
        Struct("rotation_animation", INCLUDE=anim_src_func_per_pha_sca_rot),
        QStruct("rotation_center", INCLUDE=xy_float)
        ),

    Pad(4),
    Float("zsprite_radius_scale"),

    SIZE=604
    )

rain_body = Struct("tagdata",
    Pad(36),
    reflexive("particle_types", particle_type, 8, DYN_NAME_PATH='.name'),

    SIZE=48,
    )


def get():
    return rain_def

rain_def = TagDef("rain",
    blam_header("rain"),
    rain_body,

    ext=".weather_particle_system", endian=">", tag_cls=HekTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .obje import *
from .devi import *
from .objs.mach import MachTag
from supyr_struct.defs.tag_def import TagDef

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=7)

mach_attrs = Struct("mach_attrs",
    SEnum16('type',
        'door',
        'platform',
        'gear',
        ),
    Bool16('flags',
        'pathfinding_obstable',
        'except_when_open',
        'elevator',
        ),
    float_sec('door_open_time'),  # seconds

    Pad(80),
    SEnum16('triggers_when',
        'pause_until_crushed',
        'reverse_directions'
        ),
    SInt16('elevator_node'),
    Pad(52),
    UInt32("door_open_time_ticks")
    )

mach_body = Struct("tagdata",
    obje_attrs,
    devi_attrs,
    mach_attrs,

    SIZE=804,
    )


def get():
    return mach_def

mach_def = TagDef("mach",
    blam_header('mach'),
    mach_body,

    ext=".device_machine", endian=">", tag_cls=MachTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os
import re
import tempfile

from mmap import mmap
from pathlib import Path, PureWindowsPath
from traceback import format_exc
from string import ascii_letters

from reclaimer.meta.halo_map import get_map_version, get_map_header,\
     get_tag_index, get_index_magic, get_map_magic, get_is_compressed_map,\
     decompress_map
from reclaimer.meta.wrappers.map_pointer_converter import MapPointerConverter
from reclaimer.util import is_protected_tag, int_to_fourcc, path_normalize

from supyr_struct.buffer import BytearrayBuffer, get_rawdata
from supyr_struct.util import is_path_empty


VALID_MODULE_NAME_CHARS = ascii_letters + '_' + '0123456789'


backslash_fix = re.compile(r"\\{2,}")


class HaloMap:
    map_data = None
    map_data_cache_limit = 50
    _map_cache_byte_count = 0
    _ids_of_tags_read = None

    # these are the different pieces of the map as parsed blocks
    map_header  = None
    rsrc_map    = None
    tag_index   = None
    orig_tag_index = ()  # the tag index specific to the
    #                      halo version that this map is from

    # the original tag_path of each tag in the map before any deprotection
    orig_tag_paths = None

    string_id_manager = None
    tag_index_manager = None
    map_pointer_converter = None

    # the parsed meta of the root tags in the map
    scnr_meta = None
    matg_meta = None

    # determines how to work with this map
    _filepath        = Path("")  # the filepath of the map being opened
    _decomp_filepath = Path("")  # the filepath of the map that map_data
    #                              is actually mapping. Typically only
    #                              different from filepath if the map had
    #                              to be decompressed to a different file.
    map_name        = ""
    engine          = ""
    is_resource     = False
    is_compressed   = False

    handler = None

    index_magic = 0  # the offset that halo would load the tag index
    #                  header at in virtual memory
    map_magic   = 0  # used to convert pointers in a map into file offsets.
    #                  subtract this from a pointer to convert it to an offset.
    #                      map_magic = index_magic - index_header_offset

    defs = None
    maps = None
    tag_headers = None

    tag_defs_module = ""
    tag_classes_to_load = ()
    data_extractors = {}

    resource_map_class = None

    _decomp_file_ext = ".map"

    def __init__(self, maps=None, map_data_cache_limit=None):
        self.resource_map_class = type(self)
        self.setup_defs()

        self.cache_original_tag_paths()

        self._ids_of_tags_read = set()
        if map_data_cache_limit is not None:
            self.map_data_cache_limit = HaloMap.map_data_cache_limit

        self.maps = {} if maps is None else maps

    def __del__(self):
        self.unload_map()

    @property
    def filepath(self):
        return self._filepath
    @filepath.setter
    def filepath(self, new_val):
        if not isinstance(new_val, Path):
            new_val = Path(new_val)
        self._filepath = new_val

    @property
    def decomp_filepath(self):
        return self._decomp_filepath
    @decomp_filepath.setter
    def decomp_filepath(self, new_val):
        if not isinstance(new_val, Path):
            new_val = Path(new_val)
        self._decomp_filepath = new_val

    @property
    def decomp_file_ext(self):
        return self._decomp_file_ext

    def get_writable_map_data(self):
        if not self.map_data:
            return None
        elif getattr(self.map_data, "writable", False):
            return self.map_data

        try:
            writable_map_data = get_rawdata(
                filepath=self.decomp_filepath, writable=True)
        except Exception:
            writable_map_data = None

        if not getattr(writable_map_data, "writable"):
            raise OSError("Cannot open map in write mode: %s" %
                          self.decomp_filepath)

        self.map_data.close()
        self.map_data = writable_map_data

        # need to reopen the map as a writable stream
        # and replace self.map_data with it
        return self.map_data

    # wrappers around the tag index handler
    def get_total_dir_count(self, dir=""):  return self.tag_index_manager.get_total_dir_count(dir)
    def get_total_file_count(self, dir=""): return self.tag_index_manager.get_total_file_count(dir)
    def get_dir_count(self, dir=""):  return self.tag_index_manager.get_dir_count(dir)
    def get_file_count(self, dir=""): return self.tag_index_manager.get_file_count(dir)
    def get_dir_names(self, dir=""):  return self.tag_index_manager.get_dir_names(dir)
    def get_file_names(self, dir=""): return self.tag_index_manager.get_file_names(dir)

    def rename_tag(self, tag_path, new_path):
        return self.tag_index_manager.rename_tag(tag_path, new_path)

    def rename_tag_by_id(self, tag_id, new_path):
        return self.tag_index_manager.rename_tag_by_id(tag_id, new_path)

    def rename_dir(self, curr_dir, new_dir):
        return self.tag_index_manager.rename_dir(curr_dir, new_dir)

    def print_tag_index(self, dir="", **kw):
        return self.tag_index_manager.pprint(dir, **kw)

    def print_tag_index_files(self, dir="", **kw):
        return self.tag_index_manager.pprint_files(dir, **kw)

    def walk(self, top_down=True):
        yield from self.tag_index_manager.walk(top_down)

    def setup_tag_headers(self):
        if type(self).tag_headers is not None:
            return

        tag_headers = type(self).tag_headers = {}
        for def_id in sorted(self.defs):
            if def_id in tag_headers or len(def_id) != 4:
                continue
            h_desc, h_block = self.defs[def_id].descriptor[0], [None]
            h_desc['TYPE'].parser(h_desc, parent=h_block, attr_index=0)
            tag_headers[def_id] = bytes(
                h_block[0].serialize(buffer=BytearrayBuffer(),
                                     calc_pointers=False))

    def setup_defs(self):
        assert type(self) is not HaloMap
        if type(self).defs:
            return

        print("    Loading definitions in '%s'" % self.tag_defs_module)

        type(self).defs = defs = {}
        for fcc in type(self).tag_classes_to_load:
            fcc2 = "".join(c if c in VALID_MODULE_NAME_CHARS
                           else "_" for c in fcc)
            fcc2 += "_" * ((4 - (len(fcc2) % 4)) % 4)
            try:
                # try to import the module, but ignore it if it doesnt exist
                exec("from %s.%s import get" % (self.tag_defs_module, fcc2))
                exec("defs['%s'] = get()" % fcc)
            except ImportError:
                continue
            except Exception:
                print(format_exc())

    def get_dependencies(self, meta, tag_id, tag_cls):
        return ()

    def is_indexed(self, tag_id):
        return bool(self.tag_index.tag_index[tag_id].indexed)

    def cache_original_tag_paths(self):
        tags = () if self.tag_index is None else self.tag_index.tag_index

        # record the original tag_paths so we know if they change
        self.orig_tag_paths = tuple(b.path for b in tags)

    def basic_deprotection(self):
        if self.tag_index is None:
            return

        i = 0
        found_counts = {}
        for b in self.tag_index.tag_index:
            tag_path = backslash_fix.sub(
                r'\\', b.path.replace("/", "\\")).strip().strip("\\").lower()

            name_id = (tag_path, b.class_1.enum_name)
            if is_protected_tag(tag_path):
                tag_path = "protected_%s" % i
            elif name_id in found_counts:
                tag_path = "%s_%s" % (tag_path, found_counts[name_id])
                found_counts[name_id] += 1
            else:
                found_counts[name_id] = 1

            b.path = tag_path
            i += 1

    def get_meta_descriptor(self, tag_cls):
        tagdef = self.defs.get(tag_cls)
        if tagdef is not None:
            return tagdef.descriptor[1]

    def record_map_cache_read(self, tag_id, size):
        if tag_id in self._ids_of_tags_read: return
        self._ids_of_tags_read.add(tag_id)
        self._map_cache_byte_count += size

    def map_cache_over_limit(self):
        return (self._map_cache_byte_count  >= self.map_data_cache_limit or
                len(self._ids_of_tags_read) >= self.map_data_cache_limit)

    def clear_map_cache(self):
        if not isinstance(self.map_data, mmap) or self.map_data.closed:
            return

        try:
            self.map_data.clear_cache()
        except Exception:
            pass

        self._ids_of_tags_read.clear()
        self._map_cache_byte_count = 0

    def meta_to_tag_data(self, meta, tag_cls, tag_index_ref, **kwargs):
        '''
        Changes anything in a meta data block that needs to be changed for
        it to be a working tag. This includes removing predicted_resource
        references, fetching rawdata for the bitmaps, sounds, and models,
        and byteswapping any rawdata that needs it(animations, bsp, etc).
        '''
        raise NotImplementedError()

    def inject_rawdata(self, meta, tag_cls, tag_index_ref):
        raise NotImplementedError()

    def get_meta(self, tag_id, *a, **kw):
        raise NotImplementedError()

    def get_resource_map_paths(self, maps_dir=""):
        return {}

    def is_data_extractable(self, tag_index_ref):
        return int_to_fourcc(tag_index_ref.class_1.data) in self.data_extractors

    def extract_tag_data(self, meta, tag_index_ref, **kw):
        if not self.is_data_extractable(tag_index_ref):
            return "No extractor for this type of tag."

        extractor = self.data_extractors.get(
            int_to_fourcc(tag_index_ref.class_1.data))
        kw['halo_map'] = self
        return extractor(meta, Path(PureWindowsPath(tag_index_ref.path)), **kw)

    def load_resource_maps(self, maps_dir="", map_paths=(), **kw):
        do_printout = kw.get("do_printout", False)
        detected_map_paths = self.get_resource_map_paths(maps_dir)
        if isinstance(map_paths, dict):
            for name in detected_map_paths:
                if name in map_paths:
                    detected_map_paths[name] = map_paths[name]

        map_paths = detected_map_paths
        for map_name in sorted(map_paths.keys()):
            map_path = map_paths[map_name]
            if not(self.maps.get(map_name) is None and map_path):
                continue

            new_map = None
            try:
                new_map = self.resource_map_class(self.maps)
                if do_printout:
                    print("Loading %s..." % map_name)

                new_map.load_map(map_path, **kw)
                if new_map.engine != self.engine:
                    if do_printout:
                        print("Incorrect engine for this map.")
                    self.maps.pop(new_map.map_name, None)
            except Exception:
                if do_printout:
                    print(format_exc())

                # make sure to clear out any potentially bad maps
                for name in sorted(self.maps):
                    if self.maps[name] is new_map:
                        self.maps.pop(name, None)

        return set(name for name in map_paths if not self.maps.get(name))

    def load_map(self, map_path, **kwargs):
        decompress_overwrite = kwargs.get("decompress_overwrite")
        comp_data = get_rawdata(filepath=map_path, writable=False)

        map_header = get_map_header(comp_data, True)
        if map_header is None:
            print("    Could not read map header.")
            comp_data.close()
            return

        # set self.engine early so self.decomp_file_ext will be accurate
        self.engine = get_map_version(map_header)
        self.map_name = map_header.map_name

        self.is_compressed = get_is_compressed_map(comp_data, map_header)
        if self.is_compressed:
            decomp_path = os.path.splitext(map_path)
            while decomp_path[1]:
                decomp_path = os.path.splitext(decomp_path[0])

            decomp_path = decomp_path[0] + "_DECOMP" + self.decomp_file_ext
            if not decompress_overwrite and os.path.isfile(decomp_path):
                decomp_path = os.path.join(
                    tempfile.gettempdir(), os.path.basename(decomp_path))

            print("    Decompressing to: %s" % decomp_path)
            self.map_data = decompress_map(comp_data, map_header, decomp_path)
            if comp_data is not self.map_data:
                comp_data.close()
        else:
            self.map_data = comp_data
            decomp_path = map_path

        map_header = get_map_header(self.map_data)
        tag_index  = self.orig_tag_index = get_tag_index(
            self.map_data, map_header)

        if tag_index is None:
            print("    Could not read tag index.")
            return

        self.maps[self.map_name] = self
        self.filepath        = path_normalize(map_path)
        self.decomp_filepath = path_normalize(decomp_path)
        self.map_header  = map_header
        self.index_magic = get_index_magic(map_header)
        self.map_magic   = get_map_magic(map_header)
        self.tag_index   = tag_index
        self.map_pointer_converter = MapPointerConverter()

    def unload_map(self):
        if not isinstance(getattr(self, "maps", None), dict):
            return

        keys_to_pop = list(k for k, v in self.maps.items() if v is self)
        for k in keys_to_pop:
            self.maps.pop(k, None)

        try: self.map_data.close()
        except Exception: pass

    def generate_map_info_string(self):
        if hasattr(self.map_data, '__len__'):
            decomp_size = str(len(self.map_data))
        elif (hasattr(self.map_data, 'seek') and
              hasattr(self.map_data, 'tell')):
            curr_pos = self.map_data.tell()
            self.map_data.seek(0, 2)
            decomp_size = str(self.map_data.tell())
            self.map_data.seek(curr_pos)
        else:
            decomp_size = "unknown"

        if not self.is_compressed:
            decomp_size += "(is already uncompressed)"

        map_type = self.map_header.map_type.enum_name
        if   map_type == "sp": map_type = "singleplayer"
        elif map_type == "mp": map_type = "multiplayer"
        elif map_type == "ui": map_type = "mainmenu"
        elif map_type == "shared":   map_type = "shared"
        elif map_type == "sharedsp": map_type = "shared single player"
        elif self.is_resource: map_type = "resource cache"
        elif "INVALID" in map_type:  map_type = "unknown"

        return (("""%s
Header:
    engine              == %s
    name                == %s
    build date          == %s
    type                == %s
    checksum            == %s
    decompressed size   == %s
    index header offset == %s""") %
        (self.filepath, self.engine, self.map_header.map_name,
         self.map_header.build_date, map_type, self.map_header.crc32,
         decomp_size, self.map_header.tag_index_header_offset))



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.sbsp import *
from ..common_descs import *


cluster = Struct("cluster",
    SInt16('sky'),
    SInt16('fog'),
    dyn_senum16('background_sound',
        DYN_NAME_PATH="tagdata.background_sounds_palette.STEPTREE[DYN_I].name"),
    dyn_senum16('sound_environment',
        DYN_NAME_PATH="tagdata.sound_environments_palette." +
        "STEPTREE[DYN_I].name"),
    dyn_senum16('weather',
        DYN_NAME_PATH="tagdata.weather_palettes.STEPTREE[DYN_I].name"),

    UInt16("transition_structure_bsp", VISIBLE=False),
    UInt16("first_decal_index", VISIBLE=False),
    UInt16("decal_count", VISIBLE=False),
    QStruct("unknown0",
        Float('float_0'),
        Float('float_1'),
        Float('float_2'),
        Float('float_3'),
        Float('float_4'),
        Float('float_5'),
        SIZE=24
        ),

    reflexive("predicted_resources", predicted_resource, 1024),
    reflexive("subclusters", subcluster, 4096),
    SInt16("first_lens_flare_marker_index"),
    SInt16("lens_flare_marker_count"),

    # stubbs seems to have different data here than surface indices
    Pad(12),
    reflexive("mirrors", mirror, 16, DYN_NAME_PATH=".shader.filepath"),
    reflexive("portals", portal, 128),
    SIZE=104
    )


sbsp_body = dict(sbsp_body)
sbsp_body[28] = reflexive("clusters", cluster, 8192)

fast_sbsp_body = dict(fast_sbsp_body)
fast_sbsp_body[28] = reflexive("clusters", cluster, 8192)


def get():
    return sbsp_def

sbsp_def = TagDef("sbsp",
    blam_header_stubbs("sbsp", 5),
    sbsp_body,

    ext=".scenario_structure_bsp", endian=">", tag_cls=SbspTag,
    )

fast_sbsp_def = TagDef("sbsp",
    blam_header_stubbs("sbsp", 5),
    fast_sbsp_body,

    ext=".scenario_structure_bsp", endian=">", tag_cls=SbspTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.h2.common_descs import *
from supyr_struct.defs.tag_def import TagDef

sound_classes = (
    ("projectile impact", 0),
    ("projectile detonation", 1),
    ("projectile flyby", 2),

    ("weapon fire", 4),
    ("weapon ready", 5),
    ("weapon reload", 6),
    ("weapon empty", 7),
    ("weapon charge", 8),
    ("weapon overheat", 9),
    ("weapon idle", 10),
    ("weapon melee", 11),
    ("weapon animation", 12),
    ("object impacts", 13),
    ("particle impacts", 14),

    ("unit footsteps", 18),
    ("unit dialog", 19),
    ("unit animation", 20),

    ("vehicle collision", 22),
    ("vehicle engine", 23),
    ("vehicle animation", 20),

    ("device door", 26),
    ("device machinery", 28),
    ("device stationary", 29),

    ("music", 32),
    ("ambient nature", 33),
    ("ambient machinery", 34),

    ("huge ass", 36),
    ("object looping", 37),
    ("cinematic music", 38),

    ("cortana mission", 45),
    ("cortana cinematic", 46),
    ("mission dialog", 47),
    ("cinematic dialog", 48),
    ("scripted cinematic foley", 49),
    ("game event", 50),
    ("ui", 51),
    ("test", 52),
    ("multilingual test", 53),
    )

snd__body = Struct("tagdata",
    Bool16("flags",
        "fit to adpcm blocksize",
        "split long sound into permutations",
        "always spatialize",
        "never obstruct",
        {NAME: "INTERNAL DONT TOUCH", EDITABLE:False},
        "use huge sound transmission"
        "link count to owner unit",
        "pitch range is language",
        "dont use sound class speaker flags",
        "dont use lipsync data",
        ),
    UEnum8("sound_class", *sound_classes),
    UEnum8("sample rate",
        {NAME: "khz_22", GUI_NAME: "22kHz"},
        {NAME: "khz_44", GUI_NAME: "44kHz"},
        {NAME: "khz_32", GUI_NAME: "32kHz"},
        ),
    UEnum8("encoding",
        "mono",
        "stereo",
        "codec",
        ),
    UEnum8("compression",
        "none (big endian)",
        "xbox adpcm",
        "ima adpcm",
        "none (little endian)",
        "wma",
        ),
    SInt16("playback_parameter_index"),
    SInt16("pitch_range_index"),
    SInt8("pitch_range_count"),
    SInt8("scale_index"),
    SInt8("promotion_index"),
    SInt8("custom_playback_index"),
    SInt16("extra_info_index"),
    SInt32("maximum_play_time"),
    ENDIAN="<", SIZE=20
    )


def get():
    return snd__def

snd__def = TagDef("snd!",
    h2_blam_header('snd!'),
    snd__body,

    ext=".%s" % h2_tag_class_fcc_to_ext["snd!"], endian="<"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.coll import *
from ..common_descs import *

shield = dict(shield)
shield[2] = SEnum16("shield_material_type", *materials_list)

permutation = Struct("permutation",
    ascii_str32("name"),
    ascii_str32("unknown"),
    SIZE=128
    )

region = Struct("region",
    ascii_str32("name"),
    Bool32("flags",
        "lives_until_object_dies",
        "forces_object_to_die",
        "dies_when_object_dies",
        "dies_when_object_is_damaged",
        "disappears_when_shield_is_off",
        "inhibits_melee_attack",
        "inhibits_walking",
        "forces_drop_weapon",
        "causes_head_maimed_scream",
        ),
    Pad(20),
    Float("damage_threshold"),
    QStruct("unknown_floats", *[Float("float%s" % i) for i in range(11)]),

    dependency_stubbs("destroyed_garbage", "garb"),
    dependency_stubbs("destroyed_weapon", "weap"),
    dependency_stubbs("destroyed_effect", "effe"),
    ascii_str32("unknown2"),
    Pad(28),
    reflexive("permutations", permutation, 32, DYN_NAME_PATH='.name'),
    SIZE=224
    )

material = Struct("material",
    ascii_str32("name"),
    Bool32("flags",
        "head"
        ),
    SEnum16("material_type", *materials_list),
    Pad(2),
    Float("shield_leak_percentage"),
    Float("shield_damage_multiplier"),

    Pad(16+12),
    Float("body_damage_multiplier"),
    SIZE=144
    )

coll_body = Struct("tagdata",
    Bool32("flags",
        "takes_shield_damage_for_children",
        "takes_body_damage_for_children",
        "always_shields_friendly_damage",
        "passes_area_damage_to_children",
        "parent_never_takes_body_damage_for_us",
        "only_damaged_by_explosives",
        "only_damaged_while_occupied",
        ),
    dyn_senum16("indirect_damage_material",
        DYN_NAME_PATH=".materials.materials_array[DYN_I].name"),
    Pad(2),

    body,
    shield,

    Pad(112),
    reflexive("materials", material, 32, DYN_NAME_PATH='.name'),
    reflexive("regions", region, 8, DYN_NAME_PATH='.name'),
    reflexive("modifiers", modifier, 0, VISIBLE=False),

    Pad(16),
    Struct("pathfinding_box",
        QStruct("x", INCLUDE=from_to),
        QStruct("y", INCLUDE=from_to),
        QStruct("z", INCLUDE=from_to),
        ),

    reflexive("pathfinding_spheres", pathfinding_sphere, 32),
    reflexive("nodes", node, 64, DYN_NAME_PATH='.name'),

    SIZE=664,
    )

fast_coll_body = dict(coll_body)
fast_coll_body[12] = reflexive("nodes", fast_node, 64, DYN_NAME_PATH='.name')


def get():
    return coll_def

coll_def = TagDef("coll",
    blam_header_stubbs("coll", 11),
    coll_body,

    ext=".model_collision_geometry", endian=">", tag_cls=CollTag
    )

fast_coll_def = TagDef("coll",
    blam_header_stubbs("coll", 11),  # bump this up 1 so we can tell it's stubbs
    fast_coll_body,

    ext=".model_collision_geometry", endian=">", tag_cls=CollTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


wacd_body = Struct("tagdata",
    h3_dependency("focus"),
    h3_dependency("unfocus"),
    h3_dependency("unknown_0"),
    h3_dependency("unknown_1"),
    h3_dependency("unknown_2"),
    h3_dependency("unknown_3"),
    h3_dependency("out"),
    h3_dependency("in"),
    h3_dependency("unknown_4"),
    h3_dependency("unknown_5"),
    h3_dependency("unknown_6"),
    h3_dependency("unknown_7"),
    h3_dependency("unknown_8"),
    h3_dependency("unknown_9"),
    h3_dependency("unknown_10"),
    h3_dependency("unknown_11"),
    h3_dependency("focus_in"),
    h3_dependency("focus_out"),
    h3_dependency("unknown_12"),
    h3_dependency("unknown_13"),
    h3_dependency("unknown_14"),
    h3_dependency("unknown_15"),
    h3_dependency("unknown_16"),
    h3_dependency("unknown_17"),
    h3_dependency("unknown_18"),
    h3_dependency("unknown_19"),
    h3_dependency("unknown_20"),
    h3_dependency("unknown_21"),
    h3_dependency("unknown_22"),
    h3_dependency("unknown_23"),
    h3_dependency("unknown_24"),
    h3_dependency("unknown_25"),
    h3_dependency("unknown_26"),
    h3_dependency("unknown_27"),
    h3_dependency("unknown_28"),
    h3_dependency("unknown_29"),
    ENDIAN=">", SIZE=576
    )


def get():
    return wacd_def

wacd_def = TagDef("wacd",
    h3_blam_header('wacd'),
    wacd_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["wacd"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.hek.defs.objs.mode import ModeTag

class Mod2Tag(ModeTag):

    def globalize_local_markers(self):
        tagdata = self.data.tagdata
        all_global_markers = tagdata.markers.STEPTREE
        all_global_markers_by_name = {b.name: b.marker_instances.STEPTREE
                                      for b in all_global_markers}

        for i in range(len(tagdata.regions.STEPTREE)):
            region = tagdata.regions.STEPTREE[i]
            for j in range(len(region.permutations.STEPTREE)):
                perm = region.permutations.STEPTREE[j]

                for k in range(len(perm.local_markers.STEPTREE)):
                    local_marker = perm.local_markers.STEPTREE[k]
                    global_markers = all_global_markers_by_name.get(
                        local_marker.name, None)

                    if global_markers is None or len(global_markers) >= 32:
                        all_global_markers.append()
                        all_global_markers[-1].name = local_marker.name
                        global_markers = all_global_markers[-1].marker_instances.STEPTREE
                        all_global_markers_by_name[local_marker.name] = global_markers

                    global_markers.append()
                    global_marker = global_markers[-1]

                    global_marker.region_index = i
                    global_marker.permutation_index = j
                    global_marker.node_index = local_marker.node_index
                    global_marker.rotation[:] = local_marker.rotation[:]
                    global_marker.translation[:] = local_marker.translation[:]

                del perm.local_markers.STEPTREE[:]

        # sort the markers how Halo's picky ass wants them
        name_map = {all_global_markers[i].name: i
                    for i in range(len(all_global_markers))}
        all_global_markers[:] = list(all_global_markers[name_map[name]]
                                     for name in sorted(name_map))

    def delocalize_part_nodes(self, geometry_index, part_index):
        part = self.data.tagdata.geometries.STEPTREE\
               [geometry_index].parts.STEPTREE[part_index]
        local_nodes = part.local_nodes[: part.local_node_count]

        delocalize_compressed_verts(part.compressed_vertices.STEPTREE,
                                    local_nodes)
        delocalize_uncompressed_verts(part.uncompressed_vertices.STEPTREE,
                                      local_nodes)

        part.flags.ZONER = False
        part.local_node_count = 0


def delocalize_compressed_verts(comp_verts, local_nodes):
    '''TODO: Update this function to also work on parsed vert data.'''
    local_node_ct = len(local_nodes) * 3
    # 28 is the offset to the first verts first node index
    for i in range(28, len(comp_verts), 32):
        if comp_verts[i] < local_node_ct:
            comp_verts[i] = local_nodes[comp_verts[i] // 3] * 3

        i += 1
        if comp_verts[i] < local_node_ct:
            comp_verts[i] = local_nodes[comp_verts[i] // 3] * 3


def delocalize_uncompressed_verts(uncomp_verts, local_nodes):
    '''TODO: Update this function to also work on parsed vert data.'''
    local_node_ct = len(local_nodes)
    # 57 is the offset to the least-significant half of
    # the first verts first node index(in big endian).
    for i in range(57, len(uncomp_verts), 68):
        # literally no reason to use 2 bytes for the node index since
        # a max of 63 nodes can be used, so we'll only edit the least
        # significant byte of the node indices and make it absolute.
        # if the value is 0xFF, the other byte should be 0xFF as well,
        # meaning the node is -1, which should stay as -1(no node).
        if uncomp_verts[i] < local_node_ct:
            uncomp_verts[i] = local_nodes[uncomp_verts[i]]

        i += 2
        if uncomp_verts[i] < local_node_ct:
            uncomp_verts[i] = local_nodes[uncomp_verts[i]]



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.cdmg import *
from ..common_descs import *

cdmg_body = dict(cdmg_body)
cdmg_body[5] = dict(cdmg_body[5])
cdmg_body[6] = damage_modifiers
cdmg_body[5][1] = SEnum16("category", *damage_category)
cdmg_body[5][2] = Bool32("flags",
    "does_not_hurt_owner",
    {NAME: "headshot", GUI_NAME: "causes headshots"},
    "pings_resistant_units",
    "does_not_hurt_friends",
    "does_not_ping_shields",
    "detonates_explosives",
    "only_hurts_shields",
    "causes_flaming_death",
    {NAME: "indicator_points_down",
     GUI_NAME: "damage indicator always points down"},
    "skips_shields",
    "only_hurts_one_infection_form",
    {NAME: "multiplayer_headshot",
     GUI_NAME: "causes multiplayer headshots"},
    "infection_form_pop",
    )

def get():
    return cdmg_def

cdmg_def = TagDef("cdmg",
    blam_header_stubbs('cdmg'),
    cdmg_body,

    ext=".continuous_damage_effect", endian=">"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: -DeToX-
# 	Created layout of plugin
# revision: 2		author: -DeToX-
# 	Labelled some reflexives related to raw.
# revision: 3		author: -DeToX-
# 	Labelled raw size for sounds.
# revision: 4		author: Lord Zedd
# 	Copypasta
# revision: 5		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


play_compression_codec = Struct("compression_codec",
    BytesRaw("guid", SIZE=16, EDITABLE=False),
    ENDIAN=">", SIZE=16
    )


play_external_cache_reference = Struct("external_cache_reference",
    StrLatin1("map_path", SIZE=256),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    UInt32("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=264
    )


play_raw_page = Struct("raw_page",
    SInt16("salt"),
    SInt8("flags"),
    SEnum8("compression_codec", *play_raw_page_compression_codec),
    dyn_senum16("shared_cache_index",
        DYN_NAME_PATH=".....external_cache_references.STEPTREE[DYN_I].map_path",
        WIDGET_WIDTH=32
        ),
    SInt16("unknown_page_index", SIDETIP="(Seems to be either the pages index, or -1)"),
    UInt32("block_offset"),
    UInt32("compressed_block_size"),
    UInt32("uncompressed_block_size"),
    UInt32("crc_checksum"),
    BytesRaw("entire_buffer_hash", SIZE=20, EDITABLE=False),
    BytesRaw("first_chunk_hash", SIZE=20, EDITABLE=False),
    BytesRaw("last_chunk_hash", SIZE=20, EDITABLE=False),
    SInt16("block_asset_count"),
    SInt16("unknown"),
    ENDIAN=">", SIZE=88
    )


play_raw_size_chunk = QStruct("chunk",
    # figure out what the crap this actually is
    SInt32("offset"),
    SInt32("size"),
    ENDIAN=">", SIZE=8
    )


play_raw_size = Struct("raw_size",
    SInt32("total_size"),
    h3_reflexive("chunks", play_raw_size_chunk),
    ENDIAN=">", SIZE=16
    )


play_segment = Struct("segment",
    # doing these as arrays to allow iterating over them easily in later halo's
    Array("page_indices",     SUB_STRUCT=SInt16("idx"),    SIZE=2, MIN=2, MAX=2),
    Array("segment_offsets",  SUB_STRUCT=SInt32("offset"), SIZE=2, MIN=2, MAX=2),
    Array("raw_size_indices", SUB_STRUCT=SInt16("size"),   SIZE=2, MIN=2, MAX=2),
    ENDIAN=">", SIZE=16
    )


play_body = Struct("tagdata",
    h3_reflexive("compression_codecs", play_compression_codec),
    h3_reflexive("external_cache_references", play_external_cache_reference),
    h3_reflexive("raw_pages", play_raw_page),
    h3_reflexive("raw_sizes", play_raw_size),
    h3_reflexive("segments", play_segment),
    ENDIAN=">", SIZE=60
    )


def get():
    return play_def

play_def = TagDef("play",
    h3_blam_header('play'),
    play_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["play"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.matg import *

def get():
    return matg_def

# replace the grenades reflexive with an open sauce one
matg_body = dict(matg_body)
matg_body[5] = reflexive("grenades", grenade, 4, *grenade_types_os)

matg_def = TagDef("matg",
    blam_header_os('matg', 3),
    matg_body,

    ext=".globals", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.cdmg import *

damage = Struct("damage",
    SEnum16("priority",
        "none",
        "harmless",
        {NAME: "backstab", GUI_NAME: "lethal to the unsuspecting"},
        "emp",
        ),
    SEnum16("category", *damage_category),
    Bool32("flags",
        "does_not_hurt_owner",
        {NAME: "headshot", GUI_NAME: "causes headshots"},
        "pings_resistant_units",
        "does_not_hurt_friends",
        "does_not_ping_shields",
        "detonates_explosives",
        "only_hurts_shields",
        "causes_flaming_death",
        {NAME: "indicator_points_down",
         GUI_NAME: "damage indicator always points down"},
        "skips_shields",
        "only_hurts_one_infection_form",
        {NAME: "multiplayer_headshot",
         GUI_NAME: "causes multiplayer headshots"},
        "infection_form_pop",
        "YELO_3D_instantaneous_acceleration"
        ),
    Pad(4),
    Float("damage_lower_bound"),
    QStruct("damage_upper_bound", INCLUDE=from_to),
    float_zero_to_one("vehicle_passthrough_penalty"),
    Pad(4),
    float_zero_to_one("stun"),
    float_zero_to_one("maximum_stun"),
    float_sec("stun_time"),
    Pad(4),
    QStruct("instantaneous_acceleration",
        Float("i", UNIT_SCALE=per_sec_unit_scale),
        Float("j", UNIT_SCALE=per_sec_unit_scale),
        Float("k", UNIT_SCALE=per_sec_unit_scale),
        SIDETIP="[-inf,+inf]", ORIENT="h"
        ),
    )

cdmg_body = dict(cdmg_body)
cdmg_body[5] = damage

def get():
    return cdmg_def

cdmg_def = TagDef("cdmg",
    blam_header('cdmg'),
    cdmg_body,

    ext=".continuous_damage_effect", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.hek.defs.objs.tag import HekTag


class SbspTag(HekTag):
    @staticmethod
    def point_in_front_of_plane(plane, x, y, z):
        '''
        Takes a plane ex:
        ```py
        collision_bsp = self.data.tagdata.collision_bsp.STEPTREE[0]
        collision_bsp.planes.STEPTREE[i]
        ```
        And dots the plane and coordinates and compares against d to
        see if the xyz are in front of it.
        '''
        return (x*plane.i + y*plane.j + z*plane.k) >= plane.d

    def get_leaf_index_of_point(self, x, y, z):
        '''
        Returns the leaf node the point is in in the bsp.
        If point is not in the bsp, returns None.
        '''
        if len(self.data.tagdata.collision_bsp.STEPTREE) < 1:
            raise ValueError("No collision_bsp structure found in sbsp tag.")

        collision_bsp = self.data.tagdata.collision_bsp.STEPTREE[0]
        bsp3d_nodes   = collision_bsp.bsp3d_nodes.STEPTREE
        bsp_planes    = collision_bsp.planes.STEPTREE
        if not bsp3d_nodes:
            return None

        node_index = 0
        # Go through the tree until we get a negative number (leaf or null)
        while node_index >= 0:
            node = bsp3d_nodes[node_index]
            if self.point_in_front_of_plane(bsp_planes[node.plane], x, y, z):
                node_index = node.front_child
            else:
                node_index = node.back_child

        # -1 = null (not found); otherwise it's a leaf (found)
        return None if node_index == -1 else node_index + 0x80000000

    def get_cluster_index_of_point(self, x, y, z):
        '''
        Returns the cluster the point is in in the bsp.
        If point is not in the bsp, returns None.
        '''
        leaf = self.get_leaf_index_of_point(x, y, z)
        leaves = self.data.tagdata.leaves.STEPTREE
        return leaves[leaf].cluster if leaf in range(len(leaves)) else None

    def is_point_in_bsp(self, x, y, z):
        '''Returns if given point in 3d space is inside of this BSP'''
        return self.get_leaf_index_of_point(x, y, z) is not None



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

light_volume_comment = """LIGHT VOLUME
Draws a sequence of glow bitmaps along a line. Can be used for contrail-type effects
as well as volumetric lights."""

brightness_scale_comment = """BRIGHTNESS SCALE
Fades the effect in and out with distance, viewer angle and external source."""

bitmaps_comment = """BITMAPS
Bitmap tag used to draw the light volume, repeated <count> times. Default is
'tags\\rasterizer_textures\\glow'.
Note: Sprite plates are not valid for light volumes."""

frame_animation_comment = """FRAME ANIMATION
Frames are descriptions of the light volume at a particular point in time,
interpolated by an external source. For example: A bolt of energy can be made
to stretch out and grow thinner as it is fired from a weapon."""

frame = Struct("frame",
    Pad(16),
    float_wu("offset_from_marker"),
    Float("offset_exponent"),
    float_wu("length"),

    Pad(32),
    float_wu("radius_hither"),
    float_wu("radius_yon"),
    Float("radius_exponent"),

    Pad(32),
    QStruct("tint_color_hither", INCLUDE=argb_float),
    QStruct("tint_color_yon", INCLUDE=argb_float),
    Float("tint_color_exponent"),
    Float("brightness_exponent"),
    SIZE=176
    )

mgs2_body = Struct("tagdata",
    #Light volume
    ascii_str32("attachment_marker", COMMENT=light_volume_comment),
    Bool32("flags", *blend_flags),
    Pad(16),

    #Brightness scale
    float_wu("near_fade_distance", COMMENT=brightness_scale_comment),
    float_wu("far_fade_distance"),
    float_zero_to_one("perpendicular_brightness_scale"),
    float_zero_to_one("parallel_brightness_scale"),
    SEnum16("brightness_scale_source", *function_outputs),
    Pad(22),

    #Bitmaps
    dependency("map", "bitm", COMMENT=bitmaps_comment),
    SInt16("map_sequence_index"),
    SInt16("map_count"),
    Pad(72),

    #Frame animation
    SEnum16("frame_animation_source", *function_outputs, COMMENT=frame_animation_comment),
    Pad(102),

    reflexive("frames", frame, 2),

    SIZE=332,
    )


def get():
    return mgs2_def

mgs2_def = TagDef("mgs2",
    blam_header("mgs2"),
    mgs2_body,

    ext=".light_volume", endian=">", tag_cls=HekTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from pathlib import Path
from traceback import format_exc

from arbytmap.bitmap_io import get_channel_order_by_masks,\
     get_channel_swap_mapping, swap_array_items
from supyr_struct.defs.bitmaps.dds import dds_def

__all__ = ("compile_bitmap_from_dds_files", "add_bitmap_to_bitmap_tag",
           "parse_dds_file", )


def compile_bitmap_from_dds_files(bitm_tag, dds_filepaths=()):
    for fp in dds_filepaths:
        fp = Path(fp)
        try:
            print("    %s" % fp)
            w, h, d, typ, fmt, mips, pixels = parse_dds_file(fp)
        except Exception:
            print(format_exc())
            print("    Could not load dds image")
            continue

        try:
            seq_name = fp.stem
            if "#" in seq_name:
                seq_name, _ = seq_name.split("#", 1)

            add_bitmap_to_bitmap_tag(bitm_tag, w, h, d, typ, fmt,
                                     mips, pixels, seq_name)
        except Exception:
            print(format_exc())
            print("Could not add bitmap data to bitmap tag.")


def add_bitmap_to_bitmap_tag(bitm_tag, width, height, depth, typ, fmt,
                             mip_count, new_pixels, seq_name=""):
    bitm_data = bitm_tag.data.tagdata
    sequences = bitm_data.sequences.STEPTREE
    bitmaps = bitm_data.bitmaps.STEPTREE
    seq_name = seq_name[: 31]

    if len(bitmaps) >= 2048:
        raise ValueError("Cannot add more bitmaps(max of 2048 per tag).")

    bitmaps.append()
    if not sequences or sequences[-1].sequence_name != seq_name:
        if len(sequences) >= 256:
            print("Cannot add more sequences(max of 256 per tag).")
        else:
            sequences.append()
            sequences[-1].sequence_name = seq_name
            sequences[-1].first_bitmap_index = len(bitmaps) - 1

    seq_block = sequences[-1]
    bitm_block = bitmaps[-1]
    if seq_block.sequence_name == seq_name:
        seq_block.bitmap_count += 1

    if len(bitmaps) == 1:
        if typ == "texture_2d":
            bitm_data.type.set_to("textures_2d")
        elif typ == "texture_3d":
            bitm_data.type.set_to("textures_3d")
        elif typ == "cubemap":
            bitm_data.type.set_to("cubemaps")

        if fmt == "dxt1":
            bitm_data.format.set_to("color_key_transparency")
        elif fmt == "dxt3":
            bitm_data.format.set_to("explicit_alpha")
        elif fmt == "dxt5":
            bitm_data.format.set_to("interpolated_alpha")
        elif fmt in ("r5g6b5", "a1r5g5b5", "a4r4g4b4"):
            bitm_data.format.set_to("color_16bit")
        elif fmt in ("x8r8g8b8", "a8r8g8b8", "p8_bump"):
            bitm_data.format.set_to("color_32bit")
        elif fmt in ("a8", "y8", "ay8", "a8y8"):
            bitm_data.format.set_to("monochrome")

    bitm_block.bitm_id.set_to("bitm")
    if fmt in ("dxt1", "dxt3", "dxt5"):
        bitm_block.flags.compressed = True
    bitm_block.flags.power_of_2_dim = True

    bitm_block.width = width
    bitm_block.height = height
    bitm_block.depth = depth
    bitm_block.type.set_to(typ)
    bitm_block.format.set_to(fmt)
    bitm_block.mipmaps = mip_count
    bitm_block.registration_point_x = width // 2
    bitm_block.registration_point_y = height // 2

    bitm_block.pixels_offset = len(bitm_data.processed_pixel_data.data)

    # place the pixels from the dds tag into the bitmap tag
    bitm_data.processed_pixel_data.data += new_pixels


def parse_dds_file(filepath):
    dds_tag = dds_def.build(filepath=Path(filepath))
    dds_head = dds_tag.data.header
    caps  = dds_head.caps
    caps2 = dds_head.caps2
    pixelformat = dds_head.dds_pixelformat
    pf_flags = pixelformat.flags
    dds_pixels = dds_tag.data.pixel_data
    if caps2.cubemap and not(caps2.pos_x and caps2.neg_x and
                             caps2.pos_y and caps2.neg_y and
                             caps2.pos_z and caps2.neg_z):
        raise ValueError(
            "    DDS image is malformed and does not " +
            "    contain all six necessary cubemap faces.")

    elif not dds_head.flags.pixelformat:
        raise TypeError(
            "    DDS image is malformed and does not " +
            "    contain a pixelformat structure.")

    # get the dimensions
    width = dds_head.width
    height = dds_head.height
    depth = dds_head.depth
    mip_count = max(dds_head.mipmap_count - 1, 0)
    if not caps2.volume:
        depth = 1

    # set up the flags
    fcc = pixelformat.four_cc.enum_name
    min_w = min_h = min_d = 1
    if fcc in ("DXT1", "DXT2", "DXT3", "DXT4", "DXT5"):
        min_w = min_h = 4

    bitm_format = ""
    bpp = 8  # bits per pixel
    channel_map = None

    # choose bitmap format
    if fcc == "DXT1":
        bitm_format = "dxt1"
        bpp = 4
    elif fcc in ("DXT2", "DXT3"):
        bitm_format = "dxt3"
    elif fcc in ("DXT4", "DXT5"):
        bitm_format = "dxt5"
    elif pf_flags.rgb_space:
        bitcount = pixelformat.rgb_bitcount
        bpp = 32

        if bitcount == 32:
            channel_order = get_channel_order_by_masks(
                pixelformat.a_bitmask, pixelformat.r_bitmask,
                pixelformat.g_bitmask, pixelformat.b_bitmask)

            channel_map = get_channel_swap_mapping("BGRA", channel_order)
            if pf_flags.has_alpha:
                bitm_format = "a8r8g8b8"
            elif bitcount == 32:
                bitm_format = "x8r8g8b8"

        elif bitcount in (15, 16):
            bpp = 16
            a_mask = pixelformat.a_bitmask
            r_mask = pixelformat.r_bitmask
            g_mask = pixelformat.g_bitmask
            b_mask = pixelformat.b_bitmask
            # shift the masks right until they're all the same scale
            while a_mask and not(a_mask&1): a_mask = a_mask >> 1
            while r_mask and not(r_mask&1): r_mask = r_mask >> 1
            while g_mask and not(g_mask&1): g_mask = g_mask >> 1
            while b_mask and not(b_mask&1): b_mask = b_mask >> 1

            mask_set = set((a_mask, r_mask, g_mask, b_mask))
            if mask_set == set((31, 63, 0)):
                bitm_format = "r5g6b5"
            elif mask_set == set((1, 31)):
                bitm_format = "a1r5g5b5"
            elif mask_set == set((15, )):
                bitm_format = "a4r4g4b4"

    elif pf_flags.alpha_only:
        bitm_format = "a8"

    elif pf_flags.luminance:
        if pf_flags.has_alpha:
            bitm_format = "a8y8"
        else:
            bitm_format = "y8"

    if not bitm_format:
        raise TypeError("Unknown dds image format.")

    # make sure the number of mipmaps is accurate
    face_count = 6 if caps2.cubemap else 1
    w, h, d = width, height, depth
    pixel_counts = []

    # make a list of all the pixel counts of all the mipmaps.
    for mip in range(mip_count):
        pixel_counts.append(w*h*d)
        w, h, d = (max(w//2, min_w),
                   max(h//2, min_h),
                   max(d//2, min_d))

    # see how many mipmaps can fit in the number of pixels in the dds file.
    while True:
        if (sum(pixel_counts)*bpp*face_count)//8 <= len(dds_pixels):
            break

        pixel_counts.pop(-1)

        #the mipmap count is zero and the bitmap still will
        #not fit within the space provided. Something's wrong
        if len(pixel_counts) == 0:
            raise ValueError(
                "Size of the pixel data is too small to read even " +
                "the fullsize image from. This dds file is malformed.")

    if len(pixel_counts) != mip_count:
        raise ValueError(
            "Mipmap count is too high for the number of pixels stored " +
            "in the dds file. The mipmap count has been reduced from " +
            "%s to %s." % (mip_count, len(pixel_counts)))

    mip_count = len(pixel_counts)

    # choose the texture type
    if caps2.volume:
        bitm_type = "texture_3d"
        pixels = dds_pixels

    elif caps2.cubemap:
        # gotta rearrange the mipmaps and cubemap faces
        image_count = mip_count + 1
        images = [None]*6*(image_count)
        pos = 0

        # dds images store all mips for one face next to each
        # other, and then the next set of mips for the next face.
        for face in range(6):
            w, h, d = width, height, depth
            for mip in range(image_count):
                i = mip*6 + face

                # TODO: Fix this to determine the pixel data size
                # using arbytmap's size calculation functions
                image_size = (bpp*w*h*d)//8
                images[i] = dds_pixels[pos: pos + image_size]

                w, h, d = (max(w//2, min_w),
                           max(h//2, min_h),
                           max(d//2, min_d))
                pos += image_size

        bitm_type = "cubemap"
        pixels = b''
        for image in images:
            pixels += image

    else:
        bitm_type = "texture_2d"
        pixels = dds_pixels

    if channel_map:
        # swap 32bit channels around to the format expected of bitmap tags
        pixels = bytearray(pixels)
        swap_array_items(pixels, channel_map)

    return width, height, depth, bitm_type, bitm_format, mip_count, pixels



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.meta.wrappers.halo1_map import Halo1Map
from reclaimer.shadowrun_prototype.handler import ShadowrunPrototypeHandler
from reclaimer.shadowrun_prototype.constants import sr_tag_class_fcc_to_ext
from supyr_struct.defs.frozen_dict import FrozenDict


class ShadowrunMap(Halo1Map):
    defs = None

    handler_class = ShadowrunPrototypeHandler
    tag_defs_module = "reclaimer.shadowrun_prototype.defs"
    tag_classes_to_load = tuple(sorted(sr_tag_class_fcc_to_ext.keys()))

    def setup_defs(self):
        this_class = type(self)
        if not this_class.defs:
            print("    Loading definitions in %s" %
                  self.handler_class.default_defs_path)
            this_class.defs = defs = {}

            # these imports were moved here because their defs would otherwise
            # be built when this module was imported, which is not good practice
            from reclaimer.shadowrun_prototype.defs.coll import fast_coll_def as coll_def
            from reclaimer.shadowrun_prototype.defs.sbsp import fast_sbsp_def as sbsp_def

            this_class.handler = self.handler_class(
                build_reflexive_cache=False, build_raw_data_cache=False,
                debug=2)
            this_class.defs = dict(this_class.handler.defs)
            this_class.defs["coll"] = coll_def
            this_class.defs["sbsp"] = sbsp_def
            this_class.defs = FrozenDict(this_class.defs)

        # make a shallow copy for this instance to manipulate
        self.defs = dict(self.defs)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

hud__body = Struct("tagdata",
    dependency("digits_bitmap", "bitm"),
    SInt8("bitmap_digit_width"),
    SInt8("screen_digit_width"),
    SInt8("x_offset"),
    SInt8("y_offset"),
    SInt8("decimal_point_width"),
    SInt8("colon_width"),
    SIZE=100,
    )


def get():
    return hud__def

hud__def = TagDef("hud#",
    blam_header('hud#'),
    hud__body,

    ext=".hud_number", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.ligh import LighTag
from supyr_struct.defs.tag_def import TagDef

gel_comment = """The map tints the light per-pixel of cubemap."""

lens_flare_comment = """LENS FLARE
Optional lens flare associated with this light."""

radiosity_comment = """Controls how the light affects the lightmaps (ignored for dynamic lights)."""

effect_parameters_comment = """If the light is created by an effect, it will animate itself as follows."""

ligh_body = Struct("tagdata",
    Bool32("flags",
        "dynamic",
        "no_specular",
        "dont_light_own_object",
        "supersize_in_first_person",
        "first_person_flashlight",
        "dont_fade_active_camouflage",
        ),

    #Shape
    Struct("shape",
        Float("radius"),
        QStruct("radius_modifier", INCLUDE=from_to),
        float_rad("falloff_angle"),  # radians
        float_rad("cutoff_angle"),  # radians
        Float("lens_flare_only_radius"),
        Float("cosine_falloff_angle", VISIBLE=False),
        Float("cosine_cutoff_angle", VISIBLE=False),
        Float("unknown", VISIBLE=False),
        Float("sine_cutoff_angle", VISIBLE=False),
        Pad(8),
        ),

    #Color
    Struct("color",
        Bool32("interpolation_flags", *blend_flags),
        QStruct("color_lower_bound", INCLUDE=argb_float),
        QStruct("color_upper_bound", INCLUDE=argb_float),
        Pad(12),
        ),

    #Gel
    Struct("gel_map",
        dependency("primary_cube_map", "bitm"),
        Pad(2),
        SEnum16("texture_animation_function", *animation_functions),
        float_sec("texture_animation_period"),

        dependency("secondary_cube_map", "bitm"),
        Pad(2),
        SEnum16("yaw_animation_function", *animation_functions),
        float_sec("yaw_animation_period"),
        Pad(2),
        SEnum16("roll_animation_function", *animation_functions),
        float_sec("roll_animation_period"),
        Pad(2),
        SEnum16("pitch_animation_function", *animation_functions),
        float_sec("pitch_animation_period"),
        Pad(8),
        COMMENT=gel_comment
        ),

    #Lens flare
    dependency("lens_flare", "lens", COMMENT=lens_flare_comment),
    Pad(24),

    #Radiosity
    Struct("radiosity",
        Float("intensity"),
        QStruct("color", INCLUDE=rgb_float),
        Pad(16),
        COMMENT=radiosity_comment
        ),

    #Effect parameters
    Struct("effect_parameters",
        float_sec("duration"),
        Pad(2),
        SEnum16("falloff_function", *fade_functions),
        COMMENT=effect_parameters_comment
        ),

    SIZE=352,
    )


def get():
    return ligh_def

ligh_def = TagDef("ligh",
    blam_header("ligh", 3),
    ligh_body,

    ext=".light", endian=">", tag_cls=LighTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from pathlib import Path
from traceback import format_exc

from reclaimer.halo_script.hsc import get_hsc_data_block,\
     hsc_bytecode_to_string

__all__ = ("extract_h1_scripts", )


MAX_SCRIPT_SOURCE_SIZE = 1 << 18


def extract_h1_scripts(tagdata, tag_path, **kw):
    dirpath = Path(kw.get("out_dir", "")).joinpath(
        Path(tag_path).parent, "scripts")

    overwrite = kw.get('overwrite', True)
    hsc_node_strings_by_type = kw.get("hsc_node_strings_by_type", ())

    dirpath.mkdir(exist_ok=True, parents=True)

    engine = kw.get('engine')
    if not engine and 'halo_map' in kw:
        engine = kw['halo_map'].engine

    syntax_data = get_hsc_data_block(tagdata.script_syntax_data.data)
    string_data = tagdata.script_string_data.data.decode("latin-1")
    if not syntax_data or not string_data:
        return "No script data to extract."

    already_sorted = set()
    for typ, arr in (("global", tagdata.globals.STEPTREE),
                     ("script", tagdata.scripts.STEPTREE)):
        filename_base = "%ss" % typ

        header = "; Extracted with Reclaimer\n\n"
        comments = ""
        src_file_i = 0
        global_uses  = {}
        static_calls = {}
        if typ == "global":
            sort_by = global_uses
            try:
                comments += "; scenario names(each sorted alphabetically)\n"
                comments += "\n;   object names:\n"
                for name in sorted(set(obj.name for obj in
                                       tagdata.object_names.STEPTREE)):
                    comments += ";     %s\n" % name

                comments += "\n;   trigger volumes:\n"
                for name in sorted(set(tv.name for tv in
                                       tagdata.trigger_volumes.STEPTREE)):
                    comments += ";     %s\n" % name

                comments += "\n;   device groups:\n"
                for name in sorted(set(dg.name for dg in
                                       tagdata.device_groups.STEPTREE)):
                    comments += ";     %s\n" % name

            except Exception:
                pass
        else:
            sort_by = static_calls

        try:
            sources = {}
            for i in range(len(arr)):
                # assemble source code for each function/global
                name = arr[i].name
                global_uses[name]  = set()
                static_calls[name] = set()
                sources[name] = hsc_bytecode_to_string(
                    syntax_data, string_data, i, tagdata.scripts.STEPTREE,
                    tagdata.globals.STEPTREE, typ, engine,
                    global_uses=global_uses[name],
                    hsc_node_strings_by_type=hsc_node_strings_by_type,
                    static_calls=static_calls[name])

            sorted_sources = []
            need_to_sort = sources
            while need_to_sort:
                # sort the functions/globals so dependencies come first
                next_need_to_sort = {}
                for name in sorted(need_to_sort.keys()):
                    source = need_to_sort[name]
                    if sort_by[name].issubset(already_sorted):
                        sorted_sources.append(source)
                        already_sorted.add(name)
                    else:
                        next_need_to_sort[name] = source

                if need_to_sort.keys() == next_need_to_sort.keys():
                    print("Could not sort these %ss so dependencies come first:" % typ)
                    for name in need_to_sort.keys():
                        print("\t%s" % name)
                        print("\t  Requires: ", ", ".join(sorted(sort_by[name])))
                        print()
                        sorted_sources.append(need_to_sort[name])
                    break
                need_to_sort = next_need_to_sort

            merged_sources = []
            merged_src = ""
            merged_src_len = 0
            # figure out how much data we can fit in the source file
            max_size = MAX_SCRIPT_SOURCE_SIZE - len(header) - len(comments)

            for src in sorted_sources:
                if not src:
                    continue

                src += "\n\n\n"
                # \n will be translated to \r\n, so the actual serialized string
                # length will be incremented by the number of newline characters
                src_len = len(src) + src.count("\n")

                # concatenate sources until they are too large to be compiled
                if merged_src_len + src_len >= max_size:
                    merged_sources.append(merged_src)
                    merged_src = ""
                    merged_src_len = 0

                merged_src += src
                merged_src_len += src_len

            if merged_src:
                merged_sources.append(merged_src)

            i = 0
            for out_data in merged_sources:
                # write sources to hsc files
                if len(merged_sources) > 1:
                    filename = "%s_%s.hsc" % (filename_base, i)
                else:
                    filename = "%s.hsc" % filename_base

                filepath = dirpath.joinpath(filename)
                if not overwrite and filepath.is_file():
                    continue

                # apparently the scripts use latin1 encoding, who knew....
                with filepath.open("w", encoding='latin1', newline="\r\n") as f:
                    f.write(header)
                    f.write(out_data)
                    f.write(comments)

                i += 1
        except Exception:
            return format_exc()

    # TEMPORARY CODE
    #from reclaimer.enums import TEST_PRINT_HSC_BUILT_IN_FUNCTIONS
    #TEST_PRINT_HSC_BUILT_IN_FUNCTIONS()
    # TEMPORARY CODE



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...os_v3_hek.defs.ctrl import *

#import and use the open saucified obje attrs
from .obje import *

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=8)

ctrl_body = dict(ctrl_body)
ctrl_body[0] = obje_attrs

def get():
    return ctrl_def

ctrl_def = TagDef("ctrl",
    blam_header('ctrl'),
    ctrl_body,

    ext=".device_control", endian=">", tag_cls=CtrlTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

message_index_comment = """MESSAGE INDEX
This sets which string from tags\\ui\\hud\\hud_item_messages.unicode_string_list to display."""

item_attrs = Struct('item_attrs',
    Bool32("flags",
        "always_maintains_z_up",
        "destroyed_by_explosions",
        "unaffected_by_gravity",
        ),
    SInt16("message_index", COMMENT=message_index_comment),
    SInt16("sort_order"),
    Float("scale"),
    SInt16("hud_message_value_scale"),

    Pad(18),

    SEnum16("A_in", *device_functions),
    SEnum16("B_in", *device_functions),
    SEnum16("C_in", *device_functions),
    SEnum16("D_in", *device_functions),

    Pad(164),

    dependency("material_effects", "foot"),
    dependency("collision_sound", "snd!"),

    Pad(120),

    from_to_sec("detonation_delay"),
    dependency("detonating_effect", "effe"),
    dependency("detonation_effect", "effe"),
    SIZE=396,
    )

item_body = Struct('tagdata',
    item_attrs,
    SIZE=396
    )

def get():
    return item_def

item_def = TagDef("item",
    blam_header('item', 2),
    item_body,

    ext=".item", endian=">", tag_cls=HekTag
    )
